import sys
import time

import optuna

# écriture du CSV
# premiere ligne correspond au nom des variables et aux scores
# la deuxième ligne correspond au min de chaque variable, on met un ? pour les scores
# la troisième ligne correspond au max de chaque variable, on met un ? pour les scores
# la quatrième ligne correspond au step de chaque variable, on met un ? pour les scores (cette ligne est la seule qui n'est pas obligatoire pour faire marcher le script)
#
# exemple de CSV
# Definition du problème
# var1, var2, var3, score
# 1,0,10,?
# 10,1,100,?
# 1,0.01,10,?
# Historique des essais
# 1,0.1,10,1 (score de 1 pour var1=1, var2=1, var3=1)
# 2,0.2,20,2 (score de 2 pour var1=2, var2=2, var3=2)
# 3,0.3,30,3 (score de 3 pour var1=3, var2=3, var3=3)


# Exemple: La variable 1 va de 1 à 10, avec un step de 1.
# Score est la valeur à optimiser (minimiser)
# pour faire marcher ce script il faut passer le chemin du CSV en argument
# exemple de commande pour lancer le script:
# chemin_vers_python chemin_vers_ce_script chemin_vers_le_csv


# lecture du fichier csv et controle sur les dimensions
# 0 ok
# 1 No,
# le resultat est passé sous la variable mutable tableau_out en cas de success
def lecture_fichier_csv(filename, tableau_out):
    del tableau_out[:]
    tableau = []
    try:
        with open(filename, 'r') as file:
            tableau = file.readlines()
            # suppresion du \n en fin de ligne
            for idx, element in enumerate(tableau):
                if element[-1] == "\n":
                    tableau[idx] = tableau[idx][:-1]
            file.close()
    except:
        print("erreur a la lecture de ", filename)
        return 1
    for element in tableau:
        if len(element)<2:
            continue # une seule colonne ou 0 pas possible d avoir score & variable
        tableau_out.append(element.split(","))
    if len(tableau_out) < 3:
        print("erreur a la lecture du tableau il faut au moins 3 lignes")
        return 1
    nb_element = len(tableau_out[0])
    for element in tableau_out:
        if nb_element != len(element):
            print("erreur cette ligne n a pas le bon nombre de colonne", element)
            return 1
    # on verifie que les titres des colonnes sont bien tous differents
    premiere_ligne = tableau_out[0]
    # si l entete a des doublons on dit que c'est pas bon
    if len(premiere_ligne) != len(set(premiere_ligne)):
        print("erreur nom de colonne contient des doublons")
        return 1
    return 0


# True si il y a le step d indiqué
# False Sinon
def StepDansLefile(contenu_csv):
    # le Step est eventuellement indiqué sur la 4eme ligne
    # que 3 lignes -> pas de step
    if len(contenu_csv) == 3:
        return False
    contenu_4eme_ligne = contenu_csv[3]
    # un step n a lieu que lorsque sur une colone les lignes 2 3 4 sont à ?
    if '?' not in contenu_4eme_ligne:
        return False
    # si on a que des ? sur la 4eme ligne on a pas de step
    if all(x == "?" for x in contenu_4eme_ligne):
        return False
    for idx, element in enumerate(contenu_4eme_ligne):
        if element == '?':
            if contenu_csv[1][idx] == "?" and contenu_csv[2][idx] == "?":
                return True
    return False


# recupere le domaine d inference (borne min et max par dimension) a appliquer en gerant si il y a un step de specifié
def get_distributions(contenu_du_csv):
    distributions = {}
    # on commence par chercher si on est dans le cas avec ou sans step
    avec_step = StepDansLefile(contenu_du_csv)
    # creation des distributions et des scores
    ligne_0 = contenu_du_csv[0]
    for idx, element in enumerate(ligne_0):
        if element == "":
            continue
        if contenu_du_csv[1][idx] != "?" and contenu_du_csv[2][idx] != "?":
            # nom_colonne = element
            # si borne min et max sont inversés on remet dans le bon sens, trop gentil
            if contenu_du_csv[1][idx] == '':
                continue
            if contenu_du_csv[2][idx] == '':
                continue

            borne_min = min(float(contenu_du_csv[1][idx]), float(contenu_du_csv[2][idx]))
            brone_max = max(float(contenu_du_csv[1][idx]), float(contenu_du_csv[2][idx]))
            le_step = -1
            if avec_step:
                if contenu_du_csv[3][idx] != "?":
                    le_step = float(contenu_du_csv[3][idx])
            if le_step > 0:
                distributions[element] = optuna.distributions.FloatDistribution(borne_min, brone_max, step=le_step)
            else:
                # step negatif = pas de step
                distributions[element] = optuna.distributions.FloatDistribution(borne_min, brone_max)
    return distributions


# on recupere les noms des scores à etudier
def get_nom_scores_a_etudier(contenu_du_csv):
    liste_des_scores = []
    ligne_0 = contenu_du_csv[0]
    for idx, element in enumerate(ligne_0):
        if element == "":
            continue
        if contenu_du_csv[1][idx] == "?" and contenu_du_csv[2][idx] == "?":
            liste_des_scores.append(element)
    return liste_des_scores

#on crée l'etude optuna
def create_study(contenu_csv, liste_des_noms_de_scores, distributions, param_score_previous_study, trial= False):
    #verif sur le nombre de score deja fait dans main
    if len(liste_des_noms_de_scores) == 1:
        study = optuna.create_study(sampler=optuna.samplers.TPESampler(), direction="minimize")
    else:
        directions_etude = []
        for i in range(len(liste_des_noms_de_scores)):
            directions_etude.append("minimize")
        study = optuna.create_study(sampler=optuna.samplers.MOTPESampler(), directions=directions_etude)
    ## add_trial
    for idx, elem in enumerate(param_score_previous_study):
        # je comprend pas pourquoi il faut un [0] mais sinon ca plante au bout d une dizaine d iteration
        if len(elem['scores'])==1:
            study.add_trial(
                optuna.trial.create_trial(
                params=elem['params'],
                distributions=distributions,
                values=[elem['scores'][0]]
                )
            )
        else :
            study.add_trial(
                optuna.trial.create_trial(
                    params=elem['params'],
                    distributions=distributions,
                    values=elem['scores']
                )
            )
    best_params = []
    if len(param_score_previous_study) > 0 and len(liste_des_noms_de_scores) == 1:
        best_params.append({'params':study.best_params})

    if trial == False:
        if len(param_score_previous_study) > 0 and len(liste_des_noms_de_scores) > 1:
            ## plusieurs résultats en multi voir pour comment récupérer le meilleur
            #print(study.best_trials)

            for trial in study.best_trials:
                best_params.append({'params':trial.params, 'scores': trial.values})

    trial_params = []
    ## ask
    if trial == True:
        n_trials = 1
        for _ in range(n_trials):
            # Demande de nouveaux paramètres en utilisant les distributions définies
            trial = study.ask(distributions)
            trial_params.append(trial.params)
    return trial_params, best_params

# /!\ le fichier out doit etre egal au fichier d entree il s 'agit une mise à jour
def create_ligne_from_trial_params_csv(fichier_out,trial_params, contenu_csv, liste_des_noms_de_scores):
    for j in range(len(trial_params)):
        # on ajoute au contenu_csv l'etude optuna
        ligne = []
        for i, elem in enumerate(contenu_csv[0]):
            if elem in trial_params[j].keys():
                ligne.append(str(trial_params[j][elem]))
                continue
            if elem in liste_des_noms_de_scores:
                ligne.append("?")
                continue
            else:
                ligne.append("")
        contenu_csv.append(ligne)
    try:
        with open(fichier_out, "w") as f:
            # on rempli le fichier CSV par l'etude optuna
            for i in range(len(contenu_csv)):
                ligne = ''
                for j in range(len(contenu_csv[i])):
                    ligne += contenu_csv[i][j]
                    if j+1 != len(contenu_csv[i]):
                        ligne += ","
            # Écrire dans le fichier
                f.write(str(ligne))
                f.write("\n")
            f.close()
    except Exception as error:
        print("erreur a l ecriture de ", fichier_out)
        print("An exception occurred:", error)
        return 1
    return 0

# je les les anciennes études completes (avec entrées et les scores associés)
def get_param_score_previous_study(contenu_du_csv, distributions, liste_des_noms_de_scores):
    step = StepDansLefile(contenu_du_csv)
    if step == True:
        start_ligne = 4
    else:
        start_ligne = 3
    param_score_previous_study = []
    for idx, element in enumerate(contenu_du_csv):
        # je passe le début du fichier
        if idx< start_ligne:
            continue
        # je regarde si il y a des ? et des '' sur les elements si c est le cas on ignore la ligne
        passer_la_ligne=False
        for i in range(len(contenu_du_csv[0])):
            if contenu_du_csv[0][i] in distributions.keys() or contenu_du_csv[0][i] in liste_des_noms_de_scores:
                if element[i] in ('', '?'):
                    passer_la_ligne=True
                    break
        if passer_la_ligne:
            continue
        # j ajoute les infos
        params = {}
        scores = []
        for i in range(len(contenu_du_csv[0])):
            if contenu_du_csv[0][i] in distributions.keys():
                params.update({contenu_du_csv[0][i]: float(element[i])})
            if contenu_du_csv[0][i] in liste_des_noms_de_scores:
                scores.append(float(element[i]))
        if scores != [] and params != {}:
            param_score_previous_study.append({'params': params, 'scores':scores})
    return param_score_previous_study

# renvoie 0 si ok 1 si erreur
# lecture du fichier csv (separateur ,)
# ajoute un jeu de parametre en focntions des parametres et des scores precedents obtenus
def complete_fichier_avec_proposition_optuna(filename):
    contenu_csv = []
    if 0 != lecture_fichier_csv(filename, contenu_csv):
        print("erreur a la lecture de ", filename)
        return 1
    domaine_inference_a_etudier = get_distributions(contenu_csv)
    if len(domaine_inference_a_etudier) == 0:
        print("erreur pas de domaine d inference")
        return 1
    liste_des_noms_de_scores = get_nom_scores_a_etudier(contenu_csv)
    if len(liste_des_noms_de_scores) == 0:
        print("erreur aucun score")
        return 1
    # if len(liste_des_noms_de_scores) != 1:
    #     print("erreur pour l instant je ne gere qu un seul score dans l implementation")
    #     return 1
    # print(domaine_inference_a_etudier)
    param_score_previous_study = get_param_score_previous_study(contenu_csv, domaine_inference_a_etudier, liste_des_noms_de_scores)
    trial_params = create_study(contenu_csv, liste_des_noms_de_scores, domaine_inference_a_etudier, param_score_previous_study)
    if trial_params == []:
        print("erreur pas de paramètres créés")
        return 1
    if 0!=create_ligne_from_trial_params_csv(filename,trial_params, contenu_csv, liste_des_noms_de_scores):
        print("erreur a l ecriture de ",filename)
        return 1
    return 0

def get_domain_and_data_from_previous_study(score, len_score, current_data_search_space, current_data_previous_study):
    data = []
    domain = []
    for i in range(len(current_data_search_space.domain)):
        domain.append(current_data_search_space.domain[i].name)
    if score is None and len_score is None:
        domain.append('score')
    else:
        for i in range(len_score):
            domain.append('score'+str(i))

    for i in range(len(current_data_search_space)):
        d = []
        for j in range(len(current_data_search_space.domain)):
            d.append(current_data_search_space[i][j].value)
        ## pour le paramètrage premiere iteration ajout d'un ? à la fin de la ligne
        if score is None and len_score is None:
            d.append('?')
        else:
            for i in range(len_score):
                d.append('?')
        data.append(d)
    if current_data_previous_study is not None:
        #pour eviter de rentre dans ce cas lors de la première étude dure de définir plusieurs score à l'init
        if len(current_data_previous_study) > 1:
            if len(current_data_previous_study.domain) != len(domain):
                print('Erreur les domaines de la précédente étude sont différents de l espace de recherche')
                return

        for i in range(len(current_data_previous_study)):
            d = []
            for j in range(len(current_data_previous_study.domain)):
                # dernier ligne on ajoute le score renvoyé
                if i == len(current_data_previous_study) - 1 and current_data_previous_study[i][j] == '?'  and score != None:
                    d.append(score[j-len(current_data_search_space.domain)])
                else:
                    d.append(current_data_previous_study[i][j].value)

            # on ajoute les scores car self.current_data_previous_study.domain ne contient qu'un seul score à l'init pour la première étude
            if len(current_data_previous_study) == 1 and len_score is not None and len_score > 1:
                for j in range(len_score):
                    if j != 0:
                        d.append(score[j])
            data.append(d)
    return data, domain


def get_trial_params_best_params(distributions, param_score_previous_study, domain, data, name_score, len_score, stop):
    ## si l'étude est fini on ne demande pas un nouveaux jeu de params mais juste le best params
    current_proposition_out = []
    if stop == False:
        trial_params, best_params = create_study([], name_score, distributions,
                                                              param_score_previous_study, trial=True)
        ligne = []

        for j in range(len(trial_params)):
            # on ajoute au contenu_csv l'etude optuna
            for i, elem in enumerate(domain):
                if elem in trial_params[j].keys():
                    ligne.append(str(trial_params[j][elem]))
        if len_score:
            for i in range(len_score):
                ligne.append('?')
        else:
            ligne.append('?')
        current_proposition_out = [ligne]
        data = data + [ligne]

    if stop == True:
        trial_params, best_params = create_study([], name_score, distributions,
                                                              param_score_previous_study, trial=False)

    return trial_params, best_params, data, current_proposition_out

def launch_study(score, len_score, current_data_search_space, current_data_previous_study, stop):
    data, domain = get_domain_and_data_from_previous_study(score, len_score, current_data_search_space, current_data_previous_study)

    # on créé un tableau avec les domaines et données comme demandé par optuna_multi
    new_data = [domain] + data
    distributions = get_distributions(new_data)

    param_score_previous_study = []
    best_study = []

    name_score = []
    if score is None:
        name_score.append('score')
    else:
        for i in range(len_score):
            name_score.append('score' + str(i))
    # une étude minimum pour récupérer les anciennes études
    if len(new_data) > 4:
        param_score_previous_study = get_param_score_previous_study(new_data, distributions, name_score)

    trial_params, best_params, data, current_proposition_out = get_trial_params_best_params(distributions, param_score_previous_study,
                                                                        domain, data, name_score, len_score, stop)

    ligne_best_etude = []
    if best_params != []:
        for i in range(len(param_score_previous_study)):
            for j in range(len(best_params)):
                if param_score_previous_study[i]['params'] == best_params[j]['params']:
                    best_study.append(param_score_previous_study[i])
        for j in range(len(best_study)):
            l = []
            for i, elem in enumerate(domain):
                if elem in best_study[j]['params'].keys():
                    l.append(str(best_study[j]['params'][elem]))
            for i in range(len(name_score)):
                l.append(best_study[j]['scores'][i])
            ligne_best_etude.append(l)

    return ligne_best_etude, data, domain, current_proposition_out

if __name__ == "__main__":
    if len(sys.argv)!=2:
        print("erreur attendu optuna_kh fichier_csv")
        exit(1)

    filename = str(sys.argv[1]) #"optimisation_multi_objectif_ascalon.csv" #"C:\\optuna_pour_fred\\v2\\mon_fichier.csv"
    if 0!=complete_fichier_avec_proposition_optuna(filename):
        print("erreur a la completion")
        time.sleep(10)
        exit(1)
    exit(0)


import numpy as np
import re
import os
import json
import fitz  # PyMuPDF


def load_txt_file(file_path):
    """
    Load the content of a text file and return it as a string.

    Parameters:
        file_path (str): The path to the text file to be loaded.

    Returns:
        str: The content of the text file as a single string.
    """
    try:
        with open(file_path, 'r', encoding='utf-8') as file:
            content = file.read()
        return content
    except FileNotFoundError:
        print(f"Error: The file at {file_path} was not found.")
        return ""
    except Exception as e:
        print(f"Error: An unexpected error occurred: {e}")
        return ""


def load_pdf_file(pdf_path):
    """
    Extracts and cleans the text content from a PDF using PyMuPDF.

    Args:
        pdf_path (str): Path to the PDF file.

    Returns:
        str: The extracted text content.
    """
    text = ""
    try:
        with fitz.open(pdf_path) as pdf:
            for page in pdf:
                text += page.get_text() + "\n"
    except Exception as e:
        print(f"Error reading PDF: {e}")
    return text.strip()


def chunk_string_with_overlap(text, chunk_size=300, overlap=0.5):
    """
    Splits a large string into overlapping chunks of a specified size.

    Parameters:
        text (str): The input string to be chunked.
        chunk_size (int): The number of words per chunk (default is 300).
        overlap (float): The overlap percentage between chunks (default is 0.5, i.e., 50%).

    Returns:
        list: A list of strings, each containing overlapping chunks of words.
    """
    if not 0 <= overlap < 1:
        raise ValueError("Overlap must be a float between 0 and 1 (exclusive).")

    # Split the text into a list of words
    words = text.split()

    # Calculate the overlap size in words
    overlap_size = int(chunk_size * overlap)
    step_size = chunk_size - overlap_size

    # Create overlapping chunks
    chunks = [
        ' '.join(words[i:i + chunk_size])
        for i in range(0, len(words), step_size)
        if i + chunk_size <= len(words) or i == 0  # Ensure last chunk is not too short
    ]

    return chunks


def generate_summary(model, text, max_tokens=50):
    """
    Generates a one-sentence summary of the theme of a large text using a language model (LLM).

    Parameters:
        model (Llama): The language model instance with a `.generate()` method.
        text (str): The large input text to summarize.
        max_tokens (int): The maximum number of tokens for the summary (default is 50).

    Returns:
        str: A one-sentence summary of the theme.
    """
    # Define the prompt for generating the summary
    prompt = f"""<|begin_of_text|><|start_header_id|>system<|end_header_id|>

Vous êtes un expert en analyse de document. Résumez le sujet principal du document en une seule phrase concise, sans exemple ni détails.

<|eot_id|><|start_header_id|>user<|end_header_id|>

### Texte à analyser :
{text}

### Sujet du document :<|eot_id|><|start_header_id|>assistant<|end_header_id|>
"""

    # Generate the summary using the LLM
    try:
        response = model(prompt, max_tokens=max_tokens, temperature=0, stop=["."])["choices"][0]["text"]
        return response.strip()
    except Exception as e:
        return f"Error generating summary: {e}"


def generate_dataset_on_chunk(model, embedder, chunk, threshold=0.95, max_repeat=20, safety_limit=200,
                              previous_results=None):
    """
    Generates a dataset of unique questions and answers for a given text chunk.

    Parameters:
        model: The language model used to generate questions and answers.
        embedder: A sentence embedding model used to encode questions for similarity checks.
        chunk (str): The text chunk to process.
        threshold (float): The cosine similarity threshold for considering two questions as duplicates (default: 0.95).
        max_repeat (int): The maximum number of iterations to attempt generating unique questions (default: 20).
        safety_limit (int): A hard limit on the total number of iterations to avoid infinite loops (default: 200).
        previous_results(dict): A dictionnary containing the lists of embeddings, questions and answers for the given chunk

    Returns:
        tuple: A tuple containing two lists:
            - saved_questions (list of str): The list of unique questions.
            - saved_answers (list of str): The list of answers corresponding to the questions.
    """
    # Initialize lists to store unique embeddings, questions, and answers
    if previous_results:
        saved = previous_results["embeddings"]
        saved_questions = previous_results["questions"]
        saved_answers = previous_results["answers"]
    else:
        saved = []  # Stores embeddings of unique questions
        saved_questions = []  # Stores unique questions
        saved_answers = []  # Stores answers corresponding to the unique questions
    print("Prev questions:", len(saved_questions), len(saved_answers), len(saved))
    # Initialize counters for loop control
    repeat = 0  # Tracks the number of successful iterations adding unique questions
    safety = 0  # Tracks the total number of iterations to prevent infinite loops

    # Continue generating questions until max_repeat or safety_limit is reached
    while repeat < max_repeat and safety < safety_limit:
        print(f"\rRepeated: {repeat}/{max_repeat} | Safety: {safety}/{safety_limit}", end="", flush=True)
        # Generate questions and answers using the model
        model_output = generate_questions_and_answers(model, chunk)
        questions, answers = parse_questions_and_answers(model_output)

        if questions != [] and answers != []:
            # Iterate over the generated questions and answers
            added = []
            # print(f"{len(questions)} questions generated")
            # print(f"{len(saved)} questions to compare")
            for i in range(len(questions)):
                question = questions[i]
                answer = answers[i]
                to_add = True  # Flag to indicate if the current question is unique

                # Encode the current question into embeddings
                embeddings = embedder.encode(question, show_progress_bar=False)

                # Check the similarity of the current question with previously saved questions
                for saved_embeddings in saved:
                    cos_sim = cosine_similarity(embeddings, saved_embeddings)
                    if cos_sim > threshold:  # If too similar, mark for exclusion
                        to_add = False
                        # print("---> Duplicate !", cos_sim)
                        break  # No need to check further if similarity is already too high

                # Add the question and its answer if it is unique
                if to_add:
                    saved.append(embeddings)  # Save the embeddings of the unique question
                    saved_questions.append(question)  # Save the question
                    saved_answers.append(answer)  # Save the corresponding answer

                added.append(to_add)
            # print(f"Added {sum(added)} questions to the question list ! --> {len(saved)} questions saved\n")
            if not any(added):
                repeat += 1

        # Increment the safety counter for every loop iteration
        safety += 1

    # Return the collected unique questions and their corresponding answers
    return saved_questions, saved_answers


def generate_questions_and_answers(model, text, max_tokens=4096, temperature=1, top_p=1, top_k=50, stop_tokens=None):
    """
    Generates question-answer pairs based on a provided text using a language model (LLM).

    Parameters:
        model (Llama): The language model instance with a callable API for generating text.
        text (str): The input document or excerpt to generate questions and answers from.
        max_tokens (int): Maximum tokens for the model's output (default is 4096).
        temperature (float): Sampling temperature to control randomness (default is 1).
        top_p (float): Top-p sampling to limit the token pool by probability (default is 1).
        top_k (int): Top-k sampling to limit the token pool by rank (default is 50).
        stop_tokens (list): List of stop sequences to terminate generation (default is None).

    Returns:
        str: The generated question-answer pairs.
    """
    # Define the prompt
    prompt = f"""### Contexte : On cherche à établir un jeu de questions/réponses en français sur un document, afin de finetuner un LLM sur un domaine spécifique.
Tu vas recevoir successivement des extraits de document sur lesquels tu devras générer des paires de questions/réponses pertinentes et autonomes, sans faire référence explicite au document.


### Instructions :
- Génère des questions et les réponses associées qui reflètent les informations principales de l'extrait fourni.
- La question doit être simple, précise et formulée de manière autonome. Elle ne doit pas mentionner le document.
- La réponse doit être complète et ne doit pas non plus mentionner le document.
- Respecte impérativement la structure suivante :

Q: "Question..."
R: "Réponse associée..."
Q: "Question..."
R: "Réponse associée..."


### Extrait n°6 :

"{text}"


### Assistant :
Q:"""

    try:
        # Generate output using the LLM
        response = model(prompt=prompt, max_tokens=max_tokens, temperature=temperature,
                         top_p=top_p, top_k=top_k, stop=stop_tokens)["choices"][0]["text"]
        # Extract and return the text portion of the response
        return "Q: " + response.strip()
    except Exception as e:
        return f"Error generating questions and answers: {e}"


def parse_questions_and_answers(qa_text):
    """
    Parses a string containing questions and answers into a structured list of dictionaries.

    Parameters:
        qa_text (str): The input string containing questions and answers in the format:
                       Q: "Question..."
                       R: "Answer..."

    Returns:
        list: A list of dictionaries, each containing a 'question' and 'answer' key.
    """
    # Regular expressions to match questions and answers
    question_pattern = r"Q:\s*(.+?)\s*R:"
    answer_pattern = r"R:\s*(.+?)(?:\s*Q:|$)"

    # Extract questions and answers using regex
    questions = re.findall(question_pattern, qa_text)
    answers = re.findall(answer_pattern, qa_text)

    # Ensure equal number of questions and answers
    if len(questions) != len(answers):
        return [], []
    else:
        return questions, answers


def cosine_similarity(u, v):
    dot_product = np.dot(u, v)
    norm_u = np.linalg.norm(u)
    norm_v = np.linalg.norm(v)
    return dot_product / (norm_u * norm_v)


def load_previous_results(json_path, embedder):
    """
    Load a JSON file and return a dictionary with embeddings, questions, and answers.

    Args:
        json_path (str): Path to the JSON file.
        embedder (model): Model used for embeddings.

    Returns:
        dict: A dictionary with keys 'embeddings', 'questions', and 'answers'.
    """
    # Load the JSON file
    with open(json_path, "r", encoding="utf-8") as file:
        data = json.load(file)

    # Extract questions and answers
    questions = [item["instruction"] for item in data["data"]]
    answers = [item["response"] for item in data["data"]]

    # Generate embeddings for the questions
    embeddings = embedder.encode(questions, convert_to_numpy=True)
    embeddings = [embeds for embeds in embeddings]

    # Return the result as a dictionary
    return {
        "embeddings": embeddings,
        "questions": questions,
        "answers": answers,
    }


def transform_and_save(json_path, theme):
    """
    Transforms the content of a JSON file to include a theme and overwrites the file.

    Args:
        json_path (str): Path to the JSON file to transform.
        theme (str): The theme to associate with the data.

    Returns:
        None
    """
    # Load the existing JSON data
    with open(json_path, "r", encoding="utf-8") as file:
        json_data = json.load(file)

    # Transform the data
    transformed = {
        "theme": theme,
        "data": json_data
    }

    # Save the transformed data back to the same file
    with open(json_path, "w", encoding="utf-8") as file:
        json.dump(transformed, file, ensure_ascii=False, indent=4)




import os
import sys
from Orange.data import Table, Domain, StringVariable,ContinuousVariable
from AnyQt.QtWidgets import QApplication
from AnyQt.QtCore import QEventLoop

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\","/"):
     from Orange.widgets.orangecontrib.AAIT.widgets.OWTranslation import OWTranslation
     from Orange.widgets.orangecontrib.AAIT.widgets.OWModel_HelsinkiFrEn import OWModel_HelsinkiFrEn
     from Orange.widgets.orangecontrib.AAIT.audit_widget.SignalReceiver import SignalReceiver
     from Orange.widgets.orangecontrib.AAIT.audit_widget.widgets_model import check_model_Transformers
else:
     from orangecontrib.AAIT.widgets.OWTranslation import OWTranslation
     from orangecontrib.AAIT.audit_widget.SignalReceiver import SignalReceiver
     from orangecontrib.AAIT.widgets.OWModel_HelsinkiFrEn import OWModel_HelsinkiFrEn
     from orangecontrib.AAIT.audit_widget.widgets_model import check_model_Transformers

def check_translation(translation, model, receiver):
     text_var = StringVariable("content")
     domain = Domain([], metas=[text_var])
     table = Table.from_list(domain, [["Quelle est la capital de la France ?"]])

     translation.set_model(model)
     translation.set_data(table)
     # Création d'une boucle d'attente pour simuler un chargement bloquant
     loop = QEventLoop()
     # Connecte la fin du chargement pour quitter la boucle
     def on_finish():
         loop.quit()
     translation.thread.finish.connect(on_finish)
     loop.exec_()
     translation.Outputs.data.send = receiver.receive_data
     translation.Outputs.data.send(translation.result)
     data = receiver.received_data
     if len(data) > 0:
         if len(data[0]['Translation'].value) > 0:
             print("-- Sentence: Quelle est la capital de la France ?")
             print(f"-- Translation: {data[0]['Translation'].value}")
             return 0
         else:
             return 1
     else:
         return 1

def check_widget_traduction():
    print("\n##### Translation - beginning audit")
    app = QApplication(sys.argv)
    helsinki = OWModel_HelsinkiFrEn()
    receiver = SignalReceiver()
    translation = OWTranslation()
    model = check_model_Transformers(helsinki, receiver, "Helsinki")
    if check_translation(translation, model, receiver) != 0:
        print("Erreur à la traduction")
        return 1
    print("##### Translation - audit finished\n")
    return 0


if __name__ == "__main__":
     check_widget_traduction()


import Orange
from packaging import version
target_version = version.parse("3.37")
current_version = version.parse(Orange.version.version)
if current_version < target_version: # Skip the file
    print("Orange version not compatible with all of AAIT functions !")


else: # Execute the file
    import os
    import tempfile
    import gc
    import sys
    from orangewidget.workflow.discovery import WidgetDiscovery
    from orangecanvas.registry import CategoryDescription
    from orangecanvas.registry.utils import category_from_package_globals
    from orangecanvas.utils.pkgmeta import get_distribution
    from orangecanvas.registry.utils import category_from_package_globals



    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\","/"):
         from Orange.widgets.orangecontrib.AAIT.utils.tools import first_time_check
         from Orange.widgets.orangecontrib.AAIT.utils import MetManagement
    else:
         from orangecontrib.AAIT.utils.tools import first_time_check
         from orangecontrib.AAIT.utils import MetManagement


    def remove_temp_file():
        """
        remove file which allows a message box to be displayed “launch of orange software”
        """
        temp_folder = tempfile.gettempdir()
        temp_file_path = os.path.join(temp_folder, 'orange_lance.txt')
        if os.path.exists(temp_file_path):
            try:
                os.remove(temp_file_path)
            except Exception as e:
                print(f"Erreur lors de la suppression du fichier temporaire : {e}")



    def find_widget_discovery_objects():
        """
        return a list of WidgetDiscovery isntance based on garbage collector parsing
        """
        gc.collect()
        objects = gc.get_objects()
        widget_discoveries = [obj for obj in objects if isinstance(obj, WidgetDiscovery)]
        return widget_discoveries

    def import_proprietary_categories():
        # Utilisation de la fonction pour obtenir les objets
        widget_discoveries = find_widget_discovery_objects()
        the_discovery = None
        for widget in widget_discoveries:
            the_discovery = widget
        if the_discovery is None:
            return
        pkgs = MetManagement.get_category_extension_to_load()
        if len(pkgs)==0:
            return
        sys.path.append(MetManagement.get_widget_extension_path())
        dist = get_distribution("Orange3")
        for pkg in pkgs:
            the_discovery.handle_category(category_from_package_globals(pkg))
        for pkg in pkgs:
            the_discovery.process_category_package(pkg, distribution=dist)

    def duplicate_widget_if_needed_exept_POW_file():
        path_to_check=os.path.dirname(__file__)+"/widgets"
        path_to_check=path_to_check.replace("\\","/")
        files_py = [f for f in os.listdir(path_to_check) if f.endswith('.py')]


        files_py_duplicated=[]
        files_py_not_duplicated=[]
        for element in files_py:
            if element[:2]=="__":
                files_py_duplicated.append(element)
                continue
            files_py_not_duplicated.append(element)
        for element in files_py_duplicated:
            if element=="__init__.py":
                continue
            file_origine=(element.split("__" )[-1])
            if not os.path.isfile(path_to_check+"/"+file_origine):
                try:
                    os.remove(path_to_check+"/"+element)
                    continue
                except Exception as e:
                    print(f"Error : {e}")
            if MetManagement.get_size(path_to_check+"/"+file_origine)!=MetManagement.get_size(path_to_check+"/"+element):
                # Paths to the two files
                file_1 = path_to_check+"/"+file_origine
                file_2 = path_to_check+"/"+element

                try:
                    # Read the content of the first file
                    with open(file_1, 'r') as source_file:
                        content = source_file.read()

                    # Write the content into the second file
                    with open(file_2, 'w') as target_file:
                        target_file.write(content)

                    print(f"The file '{file_2}' has been overwritten with the content of '{file_1}'.")
                    continue
                except FileNotFoundError as e:
                    print(f"Error: File not found. {e}")
                except PermissionError as e:
                    print(f"Error: Permission denied. {e}")
                except Exception as e:
                    print(f"An unexpected error occurred: {e}")


        list_file_to_created=[]
        list_file_to_duplicate=[]
        for extention in MetManagement.get_category_extension_to_load():
            dev=False
            prefix = ""
            extention_a_jouter=extention
            if len(extention) > 4:
                if extention[:3]=="dev":
                    prefix="__dev"
                    dev=True
                    # case not dev  and dev module required
                    if "site-packages/Orange/widgets" not in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
                        continue
            # case dev  and not dev module required
            if dev == False:
                if "site-packages/Orange/widgets"  in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
                    continue
            prefix+="__"+extention+"__"
            dir_to_check=MetManagement.get_widget_extension_path()+extention+"/"
            files_py = [f for f in os.listdir(dir_to_check) if f.endswith('.py')]
            for ffile in files_py:
                if ffile=="__init__.py":
                    continue
                if ffile[:3]!="POW":
                    if ffile in files_py_not_duplicated:
                        list_file_to_created.append(prefix+ffile)
                        list_file_to_duplicate.append(ffile)
                    continue


        # remove unusuable files
        for element in files_py_duplicated:
            if element =="__init__.py":
                continue
            if element in list_file_to_created:
                continue
            try:
                print("remove ",path_to_check + "/" + element)
                os.remove(path_to_check + "/" + element)
            except Exception as e:
                print(f"Error : {e}")

        for i in range(len(list_file_to_created)):
            #print(list_file_to_created[i]," ",list_file_to_duplicate[i])
            file_1 = path_to_check + "/" + list_file_to_duplicate[i]
            file_2 = path_to_check + "/" + list_file_to_created[i]
            if os.path.isfile(file_2):
                #print(file_2 ,"existe")
                continue

            try:
                # Read the content of the first file
                with open(file_1, 'r') as source_file:
                    content = source_file.read()

                # Write the content into the second file
                with open(file_2, 'w') as target_file:
                    target_file.write(content)

                print(f"The file '{file_2}' has been overwritten with the content of '{file_1}'.")
                continue
            except FileNotFoundError as e:
                print(f"Error: File not found. {e}")
            except PermissionError as e:
                print(f"Error: Permission denied. {e}")
            except Exception as e:
                print(f"An unexpected error occurred: {e}")

    def duplicate_POW_file():
        path_to_check=os.path.dirname(__file__)+"/widgets"
        path_to_check=path_to_check.replace("\\","/")
        files_py = [f for f in os.listdir(path_to_check) if f.endswith('.py') and f.startswith('POW')]
        # update pow file if necessary
        original_element="POW_Wfactory.py"
        original_pow_file=path_to_check+"/"+original_element
        if False==os.path.isfile(original_pow_file):
            print(original_element +"does not exist")
            return
        files_py_duplicated=[]
        for element in files_py:
            if element!=original_element:
                files_py_duplicated.append(element)
            if MetManagement.get_size(original_pow_file) != MetManagement.get_size(
                    path_to_check + "/" + element):
                # Paths to the two files
                file_1 = original_pow_file
                file_2 = path_to_check + "/" + element
                try:
                    # Read the content of the first file
                    with open(file_1, 'r') as source_file:
                        content = source_file.read()

                    # Write the content into the second file
                    with open(file_2, 'w') as target_file:
                        target_file.write(content)

                    print(f"The file '{file_2}' has been overwritten with the content of '{file_1}'.")
                    continue
                except FileNotFoundError as e:
                    print(f"Error: File not found. {e}")
                except PermissionError as e:
                    print(f"Error: Permission denied. {e}")
                except Exception as e:
                    print(f"An unexpected error occurred: {e}")
        list_file_to_created = []
        list_file_to_duplicate = []

        for extention in MetManagement.get_category_extension_to_load():
            dev = False
            suffix = ""
            if len(extention) > 4:
                if extention[:3] == "dev":
                    suffix = "__dev"
                    dev = True
                    # case not dev  and dev module required
                    if "site-packages/Orange/widgets" not in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
                        continue
            # case dev  and not dev module required
            if dev == False:
                if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
                    continue
            suffix+="__"+extention+"__"
            dir_to_check=MetManagement.get_widget_extension_path()+extention+"/"
            files_py = [f for f in os.listdir(dir_to_check) if f.endswith('.py') and f.startswith('POW')]
            for ffile in files_py:
                list_file_to_created.append(ffile[:-3]+suffix+".py")
                list_file_to_duplicate.append(original_element)
        # remove unusuable files
        for element in files_py_duplicated:
            if element in list_file_to_created:
                continue
            try:
                print("remove ",path_to_check + "/" + element)
                os.remove(path_to_check + "/" + element)
            except Exception as e:
                print(f"Error : {e}")
        for i in range(len(list_file_to_created)):
            #print(list_file_to_created[i]," ",list_file_to_duplicate[i])
            file_1 = path_to_check + "/" + list_file_to_duplicate[i]
            file_2 = path_to_check + "/" + list_file_to_created[i]
            if os.path.isfile(file_2):
                #print(file_2 ,"existe")
                continue

            try:
                # Read the content of the first file
                with open(file_1, 'r') as source_file:
                    content = source_file.read()

                # Write the content into the second file
                with open(file_2, 'w') as target_file:
                    target_file.write(content)

                print(f"The file '{file_2}' has been overwritten with the content of '{file_1}'.")
                continue
            except FileNotFoundError as e:
                print(f"Error: File not found. {e}")
            except PermissionError as e:
                print(f"Error: Permission denied. {e}")
            except Exception as e:
                print(f"An unexpected error occurred: {e}")

    # remove the file in temp folder
    remove_temp_file()
    import_proprietary_categories()
    duplicate_widget_if_needed_exept_POW_file()

    duplicate_POW_file()


import subprocess
import sys
import os
import psutil

def open_hide_terminal(command, with_qt=True, env=None):
    """
    Ouvre un nouveau terminal indépendant et exécute la commande spécifiée.

    :param command: La commande à exécuter.
    :param with_qt: Désactive l'affichage si False (utile pour les applications Qt sans affichage).
    :param hide_terminal: Masque la fenêtre du terminal si True.
    :return: Le PID du processus du terminal ouvert.
    """
    hide_terminal = True
    if env is None:
        env = dict(os.environ)
    if not with_qt:
        if sys.platform.startswith("darwin"):
            print("Warning: 'offscreen' mode may not be fully supported on Mac.")
        else:
            env['QT_QPA_PLATFORM'] = 'offscreen'
    if not with_qt:
        env['QT_QPA_PLATFORM'] = 'offscreen'

    process = None

    if sys.platform.startswith("win"):
        # je ne comprends pas pourquoi shell)=True cache le terminal
        process = subprocess.Popen(["cmd.exe", "/k", command], env=env, shell=True,
                                       creationflags=subprocess.CREATE_NEW_CONSOLE)

        # startupinfo = subprocess.STARTUPINFO()
        # startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        #
        # # Code 0 = SW_HIDE (cacher la console), mais ne gêne pas la GUI (ex : Qt)
        # startupinfo.wShowWindow = 0
        # creationflags = 0
        #
        # process = subprocess.Popen(
        #     command,
        #     env=env,
        #     startupinfo=startupinfo,
        #     creationflags=creationflags
        # )
    elif sys.platform.startswith("linux"):
        if hide_terminal:
            process = subprocess.Popen(["bash", "-c", command], env=env, stdout=subprocess.DEVNULL,
                                       stderr=subprocess.DEVNULL)
        else:
            try:
                process = subprocess.Popen(["gnome-terminal", "--", "bash", "-c", command + "; exec bash"], env=env)
            except FileNotFoundError:
                process = subprocess.Popen(["x-terminal-emulator", "-e", "bash", "-c", command + "; exec bash"],
                                           env=env)


    elif sys.platform.startswith("darwin"):
        if hide_terminal:
            if with_qt == False:
                command = "QT_QPA_PLATFORM=offscreen " + command
                # Lancer bash sans ouvrir un terminal visible
                process = subprocess.Popen(
                    ["/bin/bash", "-c", command],
                    env=env,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL
                )
            else:
                # Lancer bash sans ouvrir un terminal visible
                process = subprocess.Popen(
                    ["/bin/bash", "-c", command],
                    env=env,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL
                )
        else:
            if with_qt==False:
                command = "QT_QPA_PLATFORM=offscreen " + command
                # Sinon ouvrir Terminal.app
                process = subprocess.Popen(["osascript", "-e", f'tell app "Terminal" to do script \"{command}\"'])
            else:
                # Sinon ouvrir Terminal.app
                process = subprocess.Popen(["osascript", "-e", f'tell app "Terminal" to do script \"{command}\"'])
    if process:
        return process.pid
    else:
        raise RuntimeError("Impossible d'ouvrir le terminal.")


def open_terminal(command, with_qt=True, env=None):
    """
    Ouvre un nouveau terminal indépendant et exécute la commande spécifiée.
    :param command: La commande à exécuter.
    :param with_qt: Désactive l'affichage si False (utile pour les applications Qt sans affichage).
    :return: Le PID du processus du terminal ouvert.
    """
    if env is None:
        env = dict(os.environ)
    if not with_qt:
        if sys.platform.startswith("darwin"):
            print("Warning: 'offscreen' mode may not be fully supported on Mac.")
        else:
            env['QT_QPA_PLATFORM'] = 'offscreen'
    if not with_qt:
        env['QT_QPA_PLATFORM'] = 'offscreen'
    process = None

    if sys.platform.startswith("win"):
        process = subprocess.Popen(["cmd.exe", "/k", command], env=env,creationflags=subprocess.CREATE_NEW_CONSOLE)
    elif sys.platform.startswith("linux"):
        try:
            process = subprocess.Popen(["gnome-terminal", "--", "bash", "-c", command + "; exec bash"], env=env)
        except FileNotFoundError:
            process = subprocess.Popen(["x-terminal-emulator", "-e", "bash", "-c", command + "; exec bash"], env=env)
    elif sys.platform.startswith("darwin"):
        if with_qt == False:
            command = "QT_QPA_PLATFORM=offscreen "+command
            process = subprocess.Popen(["osascript", "-e", f'tell app "Terminal" to do script \"{command}\"'])
        else:
            process = subprocess.Popen(["osascript", "-e", f'tell app "Terminal" to do script \"{command}\"'])
    if process:
        return process.pid
    else:
        raise RuntimeError("Impossible d'ouvrir le terminal.")




def execute_command(command: str, hidden: bool = True,env=None):
    """
    Exécute une commande dans un terminal caché ou visible avec option d'attente.

    :param command: La ligne de commande à exécuter
    :param hidden: Exécuter dans un terminal caché si True, visible sinon
    :param wait: Attendre la fin de l'exécution si True, sinon continuer en parallèle
    :return: (code de retour, PID du processus ou None en cas d'échec)
    """
    try:
        startupinfo = None
        creationflags = 0
        shell = False

        # Gestion spécifique pour Windows
        if os.name == "nt":
            if hidden:
                startupinfo = subprocess.STARTUPINFO()
                startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                creationflags = subprocess.CREATE_NEW_CONSOLE
            else:
                command = f'start cmd.exe /k echo {command} & {command}'
            shell = True  # Nécessaire pour certaines commandes Windows
        if env is None:
            # Détermine la manière de lancer le processus
            process = subprocess.Popen(
                command,
                shell=shell,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
                startupinfo=startupinfo if os.name == "nt" else None,
                creationflags=creationflags if os.name == "nt" else 0,
            )
        else:
            # Détermine la manière de lancer le processus
            process = subprocess.Popen(
                command,
                shell=shell,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
                startupinfo=startupinfo if os.name == "nt" else None,
                creationflags=creationflags if os.name == "nt" else 0,
                env=env,
            )



        return 0, process.pid  # Succès avec PID

    except Exception as e:
        print(f"Erreur: {e}", file=sys.stderr)
        return 1, None  # Erreur


def kill_process(pid: int):
    """
    Tue un processus à partir de son PID.
    :param pid: L'identifiant du processus à tuer
    :return: 0 si succès, 1 en cas d'erreur
    """
    try:
        if os.name == "nt":
            subprocess.run(["taskkill", "/PID", str(pid), "/F"], check=True)
        else:
            os.kill(pid, 9)  # Signal SIGKILL
        return 0
    except Exception as e:
        print(f"Erreur lors de la suppression du processus {pid}: {e}", file=sys.stderr)
        return 1

def kill_process_tree(pid):
    try:
        parent = psutil.Process(pid)
        children = parent.children(recursive=True)
        for child in children:
            print(f"Tuer enfant PID {child.pid} ({child.name()})")
            child.kill()
        parent.kill()
        print(f"Processus principal {pid} tué.")
    except psutil.NoSuchProcess:
        print(f"Le processus {pid} n'existe pas.")
    except Exception as e:
        print(f"Erreur lors du kill du process tree : {e}")



import os
import gc
import sys
import json
from Orange.data import Table, Domain, StringVariable,ContinuousVariable
from AnyQt.QtWidgets import QApplication
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\","/"):
     from Orange.widgets.orangecontrib.AAIT.widgets.POW_Wfactory import POW_Wfactory
     from Orange.widgets.orangecontrib.AAIT.audit_widget.SignalReceiver import SignalReceiver
else:
     from orangecontrib.AAIT.widgets.POW_Wfactory import POW_Wfactory
     from orangecontrib.AAIT.audit_widget.SignalReceiver import SignalReceiver

def studies_out(optim, receiver):
     optim.Outputs.studies_out.send = receiver.receive_data
     optim.Outputs.studies_out.send(optim.current_data_previous_study)
     if len(receiver.received_data) > 0:
          if len(receiver.received_data[0]) != 4:
               return 1
     return 0

def check_out_pointer0(optim, receiver):
     optim.Outputs.out_pointer.send = receiver.receive_data
     optim.Outputs.out_pointer.send(str(id(optim)))
     if not isinstance(int(receiver.received_data), int):
          return 1
     return 0

def check_out_pointer1(optim, receiver):
     optim.Outputs.out_pointer1.send = receiver.receive_data
     optim.Outputs.out_pointer1.send(str(id(optim)))
     if  isinstance(int(receiver.received_data), (int)) == False:
          return 1
     return 0

def check_current_proposition_out(optim, receiver):
     optim.Outputs.current_proposition_out.send = receiver.receive_data
     optim.Outputs.current_proposition_out.send(optim.table)
     if len(receiver.received_data) > 0:
          if len(receiver.received_data[0]) != 4:
              return 1
     return 0

def check_widget_powfactory():
     # Def des espaces de recherches
     domain = Domain([ContinuousVariable("var"), ContinuousVariable("var1"), ContinuousVariable("var2")])
     rows = [[0, 0, 10], [10, 1, 100], [1, 0.1, 10]]
     table = Table.from_list(domain, rows)

     app = QApplication(sys.argv)
     factory = POW_Wfactory()
     factory.out_data = table
     factory.insert_data(0, table)



     return 0

if __name__ == "__main__":
     check_widget_powfactory()


import os
import gc
import sys
import json
from Orange.data import Table, Domain, StringVariable,ContinuousVariable
from AnyQt.QtWidgets import QApplication
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\","/"):
     from Orange.widgets.orangecontrib.AAIT.widgets.OWEditTable import OWEditTable
     from Orange.widgets.orangecontrib.AAIT.audit_widget.SignalReceiver import SignalReceiver
else:
     from orangecontrib.AAIT.widgets.OWEditTable import OWEditTable
     from orangecontrib.AAIT.audit_widget.SignalReceiver import SignalReceiver

def check_widget_edit_table():
     # Def des espaces de recherches
     domain = Domain([ContinuousVariable("var"), ContinuousVariable("var1"), ContinuousVariable("var2")])
     rows = [[0, 0, 10], [10, 1, 100], [1, 0.1, 10]]
     table = Table.from_list(domain, rows)
     app = QApplication(sys.argv)
     edittable = OWEditTable()
     receiver = SignalReceiver()
     edittable.set_data(table)
     edittable.Outputs.data.send = receiver.receive_data
     edittable.Outputs.data.send(edittable.updated_data)
     data = receiver.received_data
     if table.X.shape != data.X.shape:
         print("Erreur à l'édition des tables")
         return 1
     return 0

if __name__ == "__main__":
     check_widget_edit_table()


import fitz  # PyMuPDF
import numpy as np
import cv2
import os
import shutil
import ntpath
from tqdm import tqdm
from thefuzz import fuzz, process
from paddleocr import PaddleOCR


def page_to_image(page, dpi=300, clip=None):
    pix = page.get_pixmap(matrix=fitz.Matrix(dpi / 72, dpi / 72), clip=clip)  # Scale image
    img = np.frombuffer(pix.samples, dtype=np.uint8).reshape(pix.h, pix.w, pix.n)
    img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)  # Convert to OpenCV format
    return img


def fitz_to_cv2_rect(rect, dpi=300):
    scale = dpi / 72  # Convert from points to pixels
    x1, y1 = int(rect[0] * scale), int(rect[1] * scale)
    x2, y2 = int(rect[2] * scale), int(rect[3] * scale)
    return (x1, y1), (x2, y2)


def crop_cv2_bbox(image, bbox):
    """Crop a region from an image using OpenCV-style bounding box."""
    (x1, y1), (x2, y2) = bbox
    return image[y1:y2, x1:x2]


def draw_cv2_bbox(image, bbox, color=(0, 255, 0), thickness=2):
    """Draw a rectangle on the image using OpenCV-style bounding box."""
    (x1, y1), (x2, y2) = bbox
    img_copy = image.copy()
    cv2.rectangle(img_copy, (x1, y1), (x2, y2), color, thickness)
    return img_copy


def enlarge_fitz_bbox(bbox, w1, h1, w2, h2):
    x0, y0, x1, y1 = bbox
    width = x1 - x0
    height = y1 - y0

    new_x0 = x0 - w1 * width
    new_y0 = y0 - h1 * height
    new_x1 = x1 + w2 * width
    new_y1 = y1 + h2 * height
    return new_x0, new_y0, new_x1, new_y1


def enlarge_opencv_bbox(bbox, w1, h1, w2, h2):
    """
    Enlarge an OpenCV-style bounding box with 4 multipliers (left, top, right, bottom).

    :param bbox: ((x0, y0), (x1, y1))
    :param w1: multiplier for left side
    :param h1: multiplier for top side
    :param w2: multiplier for right side
    :param h2: multiplier for bottom side
    :return: Enlarged bbox in OpenCV format: ((new_x0, new_y0), (new_x1, new_y1))
    """
    (x0, y0), (x1, y1) = bbox
    width = x1 - x0
    height = y1 - y0

    new_x0 = int(x0 - w1 * width)
    new_y0 = int(y0 - h1 * height)
    new_x1 = int(x1 + w2 * width)
    new_y1 = int(y1 + h2 * height)

    return (new_x0, new_y0), (new_x1, new_y1)



def find_cartouche(page, output_png_path=None, output_pdf_path=None, threshold=90):
    # Extract all text from the page
    text = page.get_text("text")
    words = text.split()  # Split into individual words

    # Use fuzzy matching to find the closest match to "MATERIAL"
    pattern = "MATIERE/MATERIAL"
    best_match = process.extractOne(pattern, words, scorer=fuzz.token_sort_ratio)

    if best_match and best_match[1] >= threshold:  # Check if score meets threshold
        matched_word = best_match[0]
        #print(f"Found best match: '{matched_word}' with score: {best_match[1]}")

        # Search for the matched word in the page
        text_instances = page.search_for(matched_word)
        bbox = text_instances[0]
        if len(text_instances) > 1:
            print(f"CAREFUL !! {len(text_instances)} boxes containing '{pattern}' have been found !")

        # Modify bbox coordinates as per the original request
        bbox = enlarge_fitz_bbox(bbox, 0.5, 10, 5.5, 23)

        # Extract the selected bbox area and save it as a PNG
        page_cv2 = page_to_image(page, dpi=300, clip=bbox)

        # Draw a rectangle around the bbox in the PDF
        page.draw_rect(bbox, color=(1, 0, 0), width=2)  # Red rectangle

        # Save the crop as an image
        if output_png_path:
            cv2.imwrite(output_png_path, page_cv2)

        # Save the PDF with annotations
        if output_pdf_path:
            doc = fitz.open()
            doc.insert_pdf(page.parent, from_page=page.number, to_page=page.number)
            doc.save(output_pdf_path)
            doc.close()

        return page_cv2, bbox
    return None, None


def find_cell(page, pattern="MATIERE/MATERIAL", output_png_path=None, output_pdf_path=None, threshold=90):
    # Extract all text from the page
    text = page.get_text("text")
    words = text.split()  # Split into individual words

    # Use fuzzy matching to find the closest match to "MATERIAL"
    best_match = process.extractOne(pattern, words, scorer=fuzz.token_sort_ratio)

    if best_match and best_match[1] >= threshold:  # Check if score meets threshold
        matched_word = best_match[0]
        #print(f"Found best match: '{matched_word}' with score: {best_match[1]}")

        # Search for the matched word in the page
        text_instances = page.search_for(matched_word)
        bbox = text_instances[0]
        if len(text_instances) > 1:
            print(f"CAREFUL !! {len(text_instances)} boxes containing '{pattern}' have been found !")

        # Modify bbox coordinates to get the entire cell
        bbox = enlarge_fitz_bbox(bbox, 0.1, 0.5, 3, 0.4)

        # Extract the selected bbox area and save it as a PNG
        page_cv2 = page_to_image(page, dpi=300, clip=bbox)

        # Draw a rectangle around the bbox in the PDF
        page.draw_rect(bbox, color=(1, 0, 0), width=2)  # Red rectangle

        # Save the crop as an image
        if output_png_path:
            cv2.imwrite(output_png_path, page_cv2)

        # Save the PDF with annotations
        if output_pdf_path:
            doc = fitz.open()
            doc.insert_pdf(page.parent, from_page=page.number, to_page=page.number)
            doc.save(output_pdf_path)
            doc.close()

        return page_cv2
    return None


def find_best_match(crop, pictograms, labels, threshold=0.1, scales=None):
    import cv2

    if scales is None:
        scales = np.arange(0.1, 1.31, 0.1)

    crop_gray = cv2.cvtColor(crop, cv2.COLOR_BGR2GRAY)
    pictograms_gray = [cv2.cvtColor(p, cv2.COLOR_BGR2GRAY) for p in pictograms]

    highest_confidence = threshold
    best_index = -1

    for i, pictogram in enumerate(pictograms_gray):
        for scale in scales:
            # Resize template
            resized = cv2.resize(pictogram, None, fx=scale, fy=scale, interpolation=cv2.INTER_AREA)

            # Skip if resized template is larger than the crop
            if resized.shape[0] > crop_gray.shape[0] or resized.shape[1] > crop_gray.shape[1]:
                print("skip")
                continue

            result = cv2.matchTemplate(crop_gray, resized, cv2.TM_CCOEFF_NORMED)
            _, max_val, _, _ = cv2.minMaxLoc(result)

            if max_val > highest_confidence:
                highest_confidence = max_val
                best_index = i

    if best_index != -1:
        return pictograms[best_index], highest_confidence, labels[best_index]
    else:
        return None, highest_confidence, None


def find_pictogram_in_image(image, pictogram, threshold=0.7, max_scale=2.0, output_png_path=None):
    """
    Searches for the pictogram in the part image using template matching at multiple scales.

    :param image: Path to the part image where we search the pictogram.
    :param pictogram: Path to the pictogram image.
    :param threshold: Minimum match score for a valid match (0 to 1).
    :param max_scale: Maximum scale factor to stop resizing at.

    :return: A list of rectangles around the found pictograms along with their confidence scores.
    """
    # Convert images to grayscale
    image_gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    pictogram_gray = cv2.cvtColor(pictogram, cv2.COLOR_BGR2GRAY)

    rectangles = []
    confidences = []
    # Try different scales
    step = 0.1
    scale_values = np.arange(step, max_scale + step, step)
    for scale in tqdm(scale_values, desc="🔍 Scanning scales"):
        # Resize the pictogram at the current scale
        resized_pictogram = cv2.resize(pictogram_gray, None, fx=scale, fy=scale, interpolation=cv2.INTER_CUBIC)
        # Perform template matching
        result = cv2.matchTemplate(image_gray, resized_pictogram, cv2.TM_CCOEFF_NORMED)
        # Find locations where the match score is above the threshold
        match_locations = np.where(result >= threshold)
        for pt in zip(*match_locations[::-1]):  # Reverse the coordinates for x, y
            top_left = pt
            bottom_right = (top_left[0] + resized_pictogram.shape[1], top_left[1] + resized_pictogram.shape[0])
            confidence = result[top_left[1], top_left[0]]  # Get the confidence score for the match

            rectangles.append((top_left, bottom_right))
            confidences.append(confidence)

            # Draw rectangles on the part image (optional)
            cv2.rectangle(image, top_left, bottom_right, (0, 255, 0), 2)
    # Save the image with rectangles drawn around the found pictograms
    if output_png_path:
        cv2.imwrite(output_png_path, image)
    return rectangles, confidences


def merge_rectangles(rectangles, confidences, threshold=10):
    """
    Merges rectangles that are very close to each other based on a given threshold,
    while keeping the highest confidence rectangle.

    :param rectangles: List of tuples containing the coordinates of rectangles.
    :param confidences: List of confidence values corresponding to each rectangle.
    :param threshold: Distance threshold for merging (in pixels).
    :return: List of merged rectangles along with their associated confidence scores.
    """
    merged = []
    merged_confidences = []  # List to store the confidence scores of merged rectangles

    # Sort rectangles by top-left corner (to make it easier to compare)
    rectangles = sorted(zip(rectangles, confidences), key=lambda r: (r[0][0][1], r[0][0][0]))
    rectangles, confidences = zip(*rectangles)  # Unzip the sorted rectangles and confidences

    for rect, confidence in zip(rectangles, confidences):
        # If no merged rectangles or the current rectangle is far from the last merged one
        if not merged:
            merged.append(rect)
            merged_confidences.append(confidence)
        else:
            last_rect = merged[-1]
            last_confidence = merged_confidences[-1]

            # Check if the current rectangle is close enough to the last merged one
            x_distance = abs(rect[0][0] - last_rect[0][0])  # Check x distance
            y_distance = abs(rect[0][1] - last_rect[0][1])  # Check y distance

            if (x_distance <= threshold and y_distance <= threshold and
                    rect[1][0] >= last_rect[0][0] and rect[1][1] >= last_rect[0][1]):
                # Merge the rectangles (expand the bounding box) and keep the higher confidence
                new_rect = (
                    (min(last_rect[0][0], rect[0][0]), min(last_rect[0][1], rect[0][1])),
                    (max(last_rect[1][0], rect[1][0]), max(last_rect[1][1], rect[1][1]))
                )

                # Keep the rectangle with the higher confidence
                if confidence > last_confidence:
                    merged[-1] = new_rect
                    merged_confidences[-1] = confidence
            else:
                merged.append(rect)  # Add current rectangle to merged list if it's not close enough
                merged_confidences.append(confidence)

    return merged, merged_confidences


def exclude_inner_rectangles(big_rect, small_rects):
    """
    Removes rectangles that are fully inside the big_rect.

    :param big_rect: A tuple (x0, y0, x1, y1) for the reference rectangle.
    :param small_rects: A list of rectangles in format [((x0, y0), (x1, y1)), ...]
    :return: Filtered list of rectangles not fully inside big_rect.
    """
    bx0, by0, bx1, by1 = big_rect
    filtered = []
    for (sx0, sy0), (sx1, sy1) in small_rects:
        if not (bx0 <= sx0 <= sx1 <= bx1 and by0 <= sy0 <= sy1 <= by1):
            filtered.append(((sx0, sy0), (sx1, sy1)))
    return filtered


def apply_OCR(img, model_path):
    # Initialize the OCR model
    ocr = PaddleOCR(
        use_angle_cls=True,
        lang="fr",
        det_model_dir=os.path.join(model_path, "det"),
        rec_model_dir=os.path.join(model_path, "rec"),
        cls_model_dir=os.path.join(model_path, "cls"),
        show_log=False,
    )
    # Run OCR on the RGB image
    result = ocr.ocr(img, cls=True)
    return result






import importlib.util
import os
import pickle
from pathlib import Path

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import MetManagement
else:
    from orangecontrib.AAIT.utils import MetManagement


def delete_widget_cache(widget_name=None):
    """WINDOWS COMPATIBLE ONLY.
    This function deletes the cache of a widget, to force it to reload.
    It loads cache as pickle object, and deletes all lines containing the widget name.
    Be careful to put the widget name correctly, as it is case-sensitive. The widgetName is the name of file containing the widget.
    if widgetName is None, it will delete all cache.
    Args:
        widget_name(str): Name of the widget to delete cache of."""
    if os.name != "nt":
        print("only implemented on windows")
        return
    
    appdata_path = Path(os.getenv('LOCALAPPDATA', '')) / "Orange"
    registry_cache_files = []
    for file in appdata_path.rglob("*.pck"):
        if "registry-cache.pck" == file.name:
            registry_cache_files.append(file)
    
    if not registry_cache_files:
        print("No cache files found. Exiting the function.")
        return
            
    print("registry_cache_files: ", registry_cache_files)

    # Comment: Instead of deleting the line of the widget, you can delete the whole cache file.
    # registry_cache_file.unlink()
    for registry_cache_file in registry_cache_files:
        if widget_name is None:
            widget_name = "AAIT\\widgets"
        try:
            with open(registry_cache_file, "rb") as file:
                registry_cache = pickle.load(file)
        except Exception as e:
            print("Error loading registry cache:", e)

        new_dict ={key: registry_cache[key] for key in registry_cache.keys() if widget_name not in key}
        with open(registry_cache_file, "wb") as file:
            pickle.dump(new_dict, file)



def apply_modification_from_python_file(filepath_original_widget):
    """This decorator applies modifications to the class, reading from an external python file.
    This allows user to override or add new methods to the class, without modifying AAIT package.
    Args:
        filepath_original_widget (str): Path of the file where the original widget is defined
    """
    def decorator(cls):
        original_widget_name=filepath_original_widget.replace("\\","/")
        original_widget_name=os.path.basename(original_widget_name)
        # cases without modification
        if len(original_widget_name)<7:
            return cls
        file_to_load=MetManagement.get_widget_extension_path()
        # case dev
        if original_widget_name[:7]=="__dev__":
            original_widget_name=original_widget_name[5:]
        # cases without modification
        if original_widget_name[:2]!="__":
            if original_widget_name[:3]!="POW":
                return cls
            the_dir_the_file = original_widget_name.split("__")
            if len(the_dir_the_file)!=3 and len(the_dir_the_file)!=4:
                return cls
            file_to_load+=the_dir_the_file[-2]+"/"+the_dir_the_file[0]+".py"
        if file_to_load==MetManagement.get_widget_extension_path():# not POW
            original_widget_name=original_widget_name[2:]
            the_dir_the_file=original_widget_name.split("__")
            if len(the_dir_the_file)!=2:
                return cls
            file_to_load+=the_dir_the_file[0]+"/"+the_dir_the_file[1]
        if os.path.isfile(file_to_load)==False:
            return cls
        user_modifications_path=file_to_load
        #print(f"\n - Modification of |{cls.__name__}| based on |{user_modifications_path}|")
        # Load the module dynamically
        try:
            spec = importlib.util.spec_from_file_location("user_modifications", user_modifications_path)
            if spec is not None and spec.loader is not None:
                user_mods = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(user_mods)
            else:
                print("Error loading user modifications from file:", user_modifications_path)
        except Exception as e:
            print("Error loading user modifications from file:", user_modifications_path, e)
            return cls

        try:
            white_list = getattr(user_mods, "white_list")
        except Exception as e:
            print("Error getting white list from user modifications:", e)
            print("attributes:", dir(user_mods))
            raise e
        # after there is changes before nothing
        white_list += ["__init__"]
        #print("white list: ", white_list)
        for attribute in dir(user_mods):
            if attribute in white_list:
                #print("Replacing attribute:", attribute, "with value:", getattr(user_mods, attribute))
                attr_value = getattr(user_mods, attribute)

                if callable(attr_value):
                    # If overriding or adding new methods, use `MethodType` to bind them
                    # Assuming no need to differentiate between new and existing methods
                    # This works for `__init__`, or any other instance method
                    setattr(cls, attribute, attr_value)
                else:
                    # Directly set new class attribute values
                    setattr(cls, attribute, attr_value)

        return cls
    return decorator






import sys
import os
import numpy as np

import Orange.data
from Orange.data import ContinuousVariable, DiscreteVariable, StringVariable, TimeVariable, Domain, Table
from AnyQt.QtWidgets import QApplication
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output
from AnyQt.QtWidgets import QTableWidget, QTableWidgetItem, QComboBox, QPushButton
from AnyQt.QtCore import Qt


class OWEditTable(widget.OWWidget):
    name = "Edit Table"
    description = "Display and edit input data in a table format"
    icon = "icons/owedittable.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owedittable.svg"
    want_control_area = False
    priority = 1003

    class Inputs:
        data = Input("Data", Orange.data.Table)

    class Outputs:
        data = Output("Data", Orange.data.Table)

    class Error(widget.OWWidget.Error):
        unknown = widget.Msg("{}")

    @Inputs.data
    def set_data(self, data):
        """Receive data input and populate the table."""
        self.Error.unknown.clear()
        if data is not None:
            self.data = data
            self.populate_table()

    def __init__(self):
        super().__init__()
        # Set up the layout and table
        self.table_widget = QTableWidget()
        self.mainArea.layout().addWidget(self.table_widget)

        # Add a "Save Changes" button to trigger output
        self.btn_confirm = QPushButton("Confirm")
        self.btn_confirm.clicked.connect(self.save_changes_to_data)
        self.mainArea.layout().addWidget(self.btn_confirm)

        # Data Management
        self.data = None
        self.modified_data = {}
        self.updated_data = None


    def populate_table(self):
        """Fill QTableWidget with data from the Orange Table."""
        # Set row and column counts
        self.table_widget.setRowCount(len(self.data))
        self.table_widget.setColumnCount(len(list(self.data.domain)))

        # Set headers
        headers = [var.name for var in list(self.data.domain)]
        self.table_widget.setHorizontalHeaderLabels(headers)

        # Fill table with data
        for row in range(len(self.data)):
            for col, var in enumerate(list(self.data.domain)):
                value = str(self.data[row][var])  # Get value as string
                if isinstance(var, Orange.data.DiscreteVariable):
                    # Create a combo box for discrete variables
                    combo = QComboBox()
                    combo.addItems(var.values)  # Add possible values
                    combo.setCurrentText(value)  # Set current value
                    self.table_widget.setCellWidget(row, col, combo)
                    combo.currentTextChanged.connect(lambda new_value=value, r=row, c=var.name: self.store_modification_combo(r, c, new_value))
                else:
                    # Set as editable text for continuous and string variables
                    item = QTableWidgetItem(value)
                    item.setFlags(item.flags() | Qt.ItemIsEditable)  # Make cells editable
                    item.setData(Qt.UserRole, (row, var.name))  # Store the row and column as metadata
                    self.table_widget.setItem(row, col, item)
        # Connect the table to store user's modifications
        self.table_widget.itemChanged.connect(self.store_modification_cells)

        # Reset the modifications and the updated table
        self.modified_data = {}
        self.updated_data = self.data.copy()

        # Send the data through
        self.send_output()


    def store_modification_combo(self, row, col, new_value):
        """Store the modified values in a dictionary."""
        if col not in self.modified_data:
            self.modified_data[col] = {}  # Initialize with empty dict
        self.modified_data[col][row] = new_value


    def store_modification_cells(self, item):
        """Store the modified values in a dictionary for text cells."""
        row, col = item.data(Qt.UserRole)  # Retrieve row and column from metadata
        new_value = item.text()
        if col not in self.modified_data:
            self.modified_data[col] = {}  # Initialize with empty dict
        self.modified_data[col][row] = new_value


    def save_changes_to_data(self):
        """Modify original columns, create backup columns with original values, and send the modified table."""
        # Copy the data and domains (attributes, classes, metas)
        self.updated_data = self.data.copy()
        new_attributes = list(self.updated_data.domain.attributes)
        new_class_vars = list(self.updated_data.domain.class_vars)
        new_metas = list(self.updated_data.domain.metas)

        # Prepare new tables (X, Y, metas) as a copy of previous tables
        new_data_X = self.updated_data.X.copy()
        new_data_Y = self.updated_data.Y.copy()
        new_data_metas = self.updated_data.metas.copy()

        # Go through the dictionary containing the modifications
        for col_name, changes in self.modified_data.items():
            # Handle Attributes
            if col_name in [attr.name for attr in self.updated_data.domain.attributes]:
                col_idx = self.updated_data.domain.attributes.index(self.updated_data.domain[col_name])
                # Create backup column and update original
                new_data_X, new_attributes = self.modify_with_backup(
                    self.updated_data.domain, new_data_X, new_attributes, col_name, col_idx, changes
                )
            # Handle Class Variables
            elif col_name in [cls.name for cls in self.updated_data.domain.class_vars]:
                col_idx = self.updated_data.domain.class_vars.index(self.updated_data.domain[col_name])
                new_data_Y, new_class_vars = self.modify_with_backup(
                    self.updated_data.domain, new_data_Y, new_class_vars, col_name, col_idx, changes
                )
            # Handle Metas
            elif col_name in [meta.name for meta in self.updated_data.domain.metas]:
                col_idx = self.updated_data.domain.metas.index(self.updated_data.domain[col_name])
                new_data_metas, new_metas = self.modify_with_backup(
                    self.updated_data.domain, new_data_metas, new_metas, col_name, col_idx, changes
                )

        # Create a new domain and Table
        try:
            new_domain = Domain(new_attributes, new_class_vars, new_metas)
            self.updated_data = Table.from_numpy(new_domain, new_data_X, new_data_Y, new_data_metas)
            self.Error.unknown.clear()
        except Exception as e:
            self.Error.unknown(f"You cannot change the values of a column that has already been modified in another Edit Table widget. (detail: {e})")
            self.updated_data = self.data.copy()

        # Clear the error and send the output Table
        self.send_output()


    def modify_with_backup(self, domain, data, domain_list, col_name, col_idx, changes):
        """
        Modify original column values and create a backup column with original values.

        Parameters:
            domain (Orange.data.Domain): The domain of the Orange Data Table.
            data (np.array): The table to modify (X, Y or metas).
            domain_list (list): The corresponding domain list (attributes, class_vars, metas).
            col_name (str): The name of the column to process.
            col_idx (int): The index of the column in the np array.
            changes (dict): A dictionary mapping row indices to new values.

        Returns:
            data (np.array): Updated data with modifications.
            domain_list (list): Updated domain list with the backup column added.
        """
        if data.ndim == 1:
            data = data.reshape(-1, 1)
        # Get the original column
        original_col = data[:, col_idx].copy()

        # Modify the original column
        var = domain[col_name]
        if isinstance(var, TimeVariable):
            for row_idx, new_value in changes.items():
                try:
                    original_col[int(row_idx)] = var.parse(new_value)
                except ValueError as e:
                    self.Error.unknown(f"Invalid date format: {e}")
                    raise ValueError("Invalid date format")

        elif isinstance(var, ContinuousVariable):
            for row_idx, new_value in changes.items():
                try:
                    original_col[int(row_idx)] = float(new_value)
                except ValueError as e:
                    self.Error.unknown(f"Invalid continuous value: {e}")
                    raise ValueError("Invalid continuous value")

        elif isinstance(var, DiscreteVariable):
            for row_idx, new_value in changes.items():
                try:
                    original_col[int(row_idx)] = var.values.index(new_value)
                except ValueError as e:
                    self.Error.unknown(f"Value not in categories: {new_value}. {e}")
                    raise ValueError("Value not in categories")

        elif isinstance(var, StringVariable):
            original_col = original_col.astype(object)
            for row_idx, new_value in changes.items():
                original_col[int(row_idx)] = str(new_value)

        # Create a backup column and corresponding variable
        backup_col_name = f"{col_name} (Original)"
        backup_var = type(var)(backup_col_name, getattr(var, "values", None))
        data = np.column_stack((data, data[:, col_idx]))  # Add original values as backup
        domain_list.append(backup_var)  # Add backup variable to domain

        # Replace the original column with the modified one
        data[:, col_idx] = original_col
        return data, domain_list


    def send_output(self):
        """Send the modified table with duplicated columns."""
        self.Outputs.data.send(self.updated_data)
        return


if __name__ == "__main__":
    app = QApplication(sys.argv)
    obj = OWEditTable()
    obj.show()
    app.exec_()


from itertools import chain
from os import listdir, environ
from os.path import isfile, join, dirname
import unittest
from unittest import mock

from AnyQt.QtTest import QTest

from orangecanvas.registry import WidgetRegistry
from orangecanvas.config import EntryPoint
from orangewidget.workflow import widgetsscheme

from Orange.canvas.config import Config
from Orange.canvas import workflows

from Orange.widgets.tests.base import GuiTest
from Orange.widgets.tests.utils import excepthook_catch


def discover_workflows(dir):
    ows_files = [f for f in listdir(dir)
                 if isfile(join(dir, f)) and f.endswith(".ows")]
    for ows_file in ows_files:
        yield join(dir, ows_file)


def registry():
    d = Config.widget_discovery(WidgetRegistry())
    d.run(Config.widgets_entry_points())
    return d.registry


@unittest.skipIf(environ.get("SKIP_EXAMPLE_WORKFLOWS", False),
                 "Example workflows inflate coverage")
class TestWorkflows(GuiTest):
    def test_scheme_examples(self):
        """
        Test if Orange workflow examples can be opened. Examples in canvas
        and also those placed "workflows" subfolder.
        GH-2240
        """
        reg = registry()
        test_workflows = chain(
            discover_workflows(dirname(workflows.__file__)),
            discover_workflows(join(dirname(__file__), "workflows"))
        )
        for ows_file in test_workflows:
            new_scheme = widgetsscheme.WidgetsScheme()
            new_scheme.widget_manager.set_creation_policy(
                new_scheme.widget_manager.Immediate
            )
            new_scheme.signal_manager.pause()
            with open(ows_file, "rb") as f:
                try:
                    with excepthook_catch(raise_on_exit=True):
                        new_scheme.load_from(f, registry=reg)
                except Exception as e:
                    self.fail("Old workflow '{}' could not be loaded\n'{}'".
                              format(ows_file, str(e)))
                finally:
                    new_scheme.clear()
                    new_scheme.deleteLater()
                    del new_scheme
                    QTest.qWait(0)

    def test_examples_order(self):
        ep_first = EntryPoint(
            '!Testname', 'orangecontrib.any_addon.tutorials', '')
        ep_last = EntryPoint(
            'exampletutorials', 'orangecontrib.other_addon.tutorials', '')

        def entry_points(*_, **__):
            return ep_first, ep_last

        with mock.patch("Orange.canvas.config.entry_points", entry_points):
            ep_names = [point.name for point in Config.examples_entry_points()]
        self.assertLess(ep_names.index(ep_first.name),
                        ep_names.index("000-Orange3"))
        self.assertLess(ep_names.index("000-Orange3"),
                        ep_names.index(ep_last.name))



import argparse
import gc
import os

import Orange


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import (MetManagement, shared_variables)
else:
    from orangecontrib.AAIT.utils import (MetManagement, shared_variables)


def get_ow_file_widget():
    """
    This function retrieves the Orange Canvas file widget associated with the main window.
    It iterates through the objects in the garbage collector and identifies the main window,
    then prints information about the associated widgets.
    """
    le_self = None  # Variable to store the main window instance

    # Iterate through the objects in the garbage collector
    for obj in gc.get_objects():
        if isinstance(obj, Orange.canvas.mainwindow.MainWindow):
            le_self = obj
    return le_self


def get_duplicate_names(only_aait=False):
    """
    This function collects and compares widget names for every Orange Widget Scheme (OWS) in the Orange Canvas.
    It prints information about the widgets and their associations with OWS.
    It identifies duplicate widget names across different OWS and returns a list of tuples containing OWS names
    and sets of duplicate widget names.

    Parameters:
    - only_aait: If True, targets only widgets with category starting with "aait". If False, includes all widgets.

    Returns:
    A list of tuples containing OWS names and sets of duplicate widget names.
    """
    ows_list = []  # List to store OWS names
    widgets_lists = []  # List to store lists of widgets for each OWS

    print("Collecting all widgets for every OWS")
    for obj in gc.get_objects():
        if isinstance(obj, Orange.canvas.mainwindow.MainWindow):
            ows_name = obj.current_document().path().split("/")[-1]
            ows_list.append(ows_name)
            widgets_list = []

            # Extract information about widgets for the current OWS
            item_dict = obj.scheme_widget._SchemeEditWidget__scheme.widget_manager._OWWidgetManager__item_for_node
            for item in item_dict.items():
                #widget = item[1].widget  # The widget is the 2nd element of the Tuple
                widget_name = item[0].title
                # print("  OWS :", ows_name, "ID :", id(obj))
                # print("  Widget :", widget)
                # print("  Widget ID:", id(widget))
                # print("  Widget name:", widget_name)
                category = item[0].description.category

                # Targets only aait widgets or includes all widgets based on the 'only_aait' parameter
                if category[0:9] == "dev_workflow" or not only_aait:
                    widgets_list.append(widget_name)

            widgets_lists.append(widgets_list)

    all_duplicates = []  # List to store tuples of OWS names and sets of duplicate widget names
    n_ows = len(ows_list)

    # Compare the lists of widgets to find duplicates
    for i in range(n_ows):
        for j in range(i + 1, n_ows):
            ows_1 = ows_list[i]
            ows_2 = ows_list[j]
            widgets_list_1 = widgets_lists[i]
            widgets_list_2 = widgets_lists[j]
            duplicates = set(widgets_list_1) & set(widgets_list_2)

            if duplicates:
                all_duplicates.append((ows_1, ows_2, duplicates))

    return all_duplicates


def get_current_ows_gui(arg_self):
    """
    Returns the Orange Widget Scheme (OWS) object containing the specified widget.

    If the OWS hasn't been named and saved, it forces a save-as operation.

    Parameters:
    - arg_self: The widget for which the OWS is to be retrieved.

    Returns:
    The OWS object containing the widget, or None if not found.
    """
    # print("sys argv0", sys.argv[0])
    # print("arg_self current ows", arg_self)
    widget_id = id(arg_self)
    # print("----> Created Widget ID :", widget_id)

    # Loop over ALL instantiated objects
    for obj in gc.get_objects():
        # Stop whenever a MainWindow is found (OWS)
        if isinstance(obj, Orange.canvas.mainwindow.MainWindow):
            # Dictionary containing all the existing widgets (inside Tuples)
            item_dict = obj.scheme_widget._SchemeEditWidget__scheme.widget_manager._OWWidgetManager__item_for_node

            for item in item_dict.items():
                # The widget (object) is the 2nd element of Tuple
                widget = item[1].widget

                if id(widget) == widget_id:
                    ows_path = obj.current_document().path()

                    # If the OWS hasn't been named / saved, force save-as
                    if not ows_path:
                        MetManagement.force_save_as(arg_self, target_canvas_id=id(obj))

                    return obj

    return None


def get_current_ows_name_no_gui(arg_self):
    """
    Returns the name of the Orange Widget Scheme (OWS) associated with the specified argument namespace.

    Parameters:
    - arg_self: The argument namespace for which the OWS name is to be retrieved.

    Returns:
    The name of the associated OWS or an empty string if not found.
    """
    for obj in gc.get_objects():
        if isinstance(obj, argparse.Namespace):
            ows_path = obj.file
            return ows_path

    return ""


def setup_shared_variables(arg_self):
    """
    Setup for shared variables. Should be put at the beginning of __init__ of every aait widget.

    Parameters:
    - arg_self: The instance of the widget where shared variables are being set up.
    """
    # Case where coding a widget outside Orange
    if hasattr(Orange, "canvas") == False:
        arg_self.current_ows = shared_variables.current_ows
        shared_variables.ptr_current_canvas_main = 0
        return

    if hasattr(Orange.canvas, "mainwindow"):
        ows = get_current_ows_gui(arg_self)

        # Retrieve the path of the current Orange Widget Scheme (OWS)
        try:
            ows_path = ows.current_document().path()
        except:
            ows_path = None

        arg_self.current_ows = ows_path
        shared_variables.current_ows = ows_path
        shared_variables.ptr_current_canvas_main = id(ows)
        # Add the [ows_path, id(ows)] pair to the list if not already present
        if [ows_path, id(ows)] not in shared_variables.vect_doc_ptr:
            shared_variables.vect_doc_ptr.append([ows_path, id(ows)])

        # print("all ows", shared_variables.vect_doc_ptr)
        # print("/!\\ shared_variables current doc", shared_variables.current_ows)
    else:
        # Retrieve the path of the current Orange Widget Scheme (OWS) in the absence of GUI
        ows_path = get_current_ows_name_no_gui(arg_self)

        arg_self.current_ows = ows_path
        shared_variables.current_ows = ows_path
        shared_variables.ptr_current_canvas_main = 0



import os
import sys
import re
import Orange.data


from Orange.classification.rules import _RuleClassifier
from Orange.widgets import widget
from Orange.widgets.utils.annotated_data import (create_annotated_table,
                                                 ANNOTATED_DATA_SIGNAL_NAME)
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import Input, Output
from Orange.data import Table, Domain,  StringVariable

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
from Orange.widgets.settings import Setting
from AnyQt.QtWidgets import QPushButton, QCheckBox

@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWConcatRules(widget.OWWidget):
    name = "Intersect CN2 Rules"
    description = "Concat CN2 Rules from data."
    icon = "icons/owCN2_intersect_rules.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owCN2_intersect_rules.svg"
    priority = 1145
    keywords = "cn2 rule concat"
    want_control_area = False
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owconcatrules.ui")

    class Inputs:
        rules = Input("Rules", Orange.data.Table)
        new_rules = Input("New rules", Orange.data.Table)

    class Outputs:
        concat_rules =Output("Concat Rules", Orange.data.Table)

    strauto: str = Setting('False')



    @Inputs.rules
    def set_rules(self, data):
        new_data = self.fake_table_from_min_max(data)
        self.rules_data = new_data
        self.Outputs.concat_rules.send(None)
        if self.strauto == 'True' or self.strauto == 'Middle':
            self.run()
            if self.strauto == 'Middle':
                self.strauto = 'False'
                self.checkbox_interface.setChecked(False)

    @Inputs.new_rules
    def set_new_rules(self, data):
        new_data = self.fake_table_from_min_max(data)
        self.new_rules_data = new_data
        self.Outputs.concat_rules.send(None)
        if self.strauto == 'True' or self.strauto == 'Middle':
            self.run()
            if self.strauto == 'Middle':
                self.strauto = 'False'
                self.checkbox_interface.setChecked(False)

    def __init__(self):
        super().__init__()
        # Set the fixed width and height of the widget
        self.setFixedWidth(470)
        self.setFixedHeight(300)

        # Load the user interface file
        uic.loadUi(self.gui, self)
        self.rules_data = None
        self.new_rules_data = None
        self.btn_run = self.findChild(QPushButton, "pushButton")
        self.btn_run.clicked.connect(self.run)

        self.btn_run_first_rule = self.findChild(QPushButton, "pushButton_2")
        self.btn_run_first_rule.clicked.connect(self.send_first_rule)
        self.checkbox_interface = self.findChild(QCheckBox, 'checkBox')
        if self.strauto == 'Middle':
            self.strauto = 'False'
        if self.strauto == 'False':
            self.checkbox_interface.setChecked(False)
        if self.strauto == 'True':
            self.checkbox_interface.setChecked(True)
        self.checkbox_interface.stateChanged.connect(self.on_checkbox_toggled)
        self.post_initialized()

    def send_first_rule(self):
        self.error("")
        if self.rules_data is  None:
            self.error("you need rule data")
            return
        self.Outputs.concat_rules.send(self.rules_data)

    def fake_table_from_min_max(self,input_data):
        if input_data is None:
            return input_data
        if "regle" in input_data.domain:
            return input_data
        if "Feature" in input_data.domain and "Min." in input_data.domain and "Max." in input_data.domain:
            str_rules = ""
            idx_min = input_data.domain.index("Min.")
            idx_max = input_data.domain.index("Max.")
            idx_feature = input_data.domain.index("Feature")
            for line in input_data:
                if line[idx_min] == "?":
                    continue
                if line[idx_max] == "?":
                    continue
                to_add = str(line[idx_feature]) + " >= " + str(line[idx_min])
                to_add += " and " + str(line[idx_feature]) + " <= " + str(line[idx_max])

                if str_rules != "":
                    str_rules += " and "
                str_rules += to_add
            meta = [[str_rules]]  # 1 ligne, 1 colonne meta

            domain = Domain([], metas=[StringVariable("regle")])
            data = [[]]  # Ajoute une ligne vide pour correspondre aux métas

            table = Table.from_numpy(domain, data, metas=meta)
            return table

        if "name" in input_data.domain and "min" in input_data.domain and "max" in input_data.domain:
            str_rules = ""
            idx_min = input_data.domain.index("min")
            idx_max = input_data.domain.index("max")
            idx_feature = input_data.domain.index("name")
            for line in input_data:
                if line[idx_min] == "?":
                    continue
                if line[idx_max] == "?":
                    continue
                to_add = str(line[idx_feature]) + " >= " + str(line[idx_min])
                to_add += " and " + str(line[idx_feature]) + " <= " + str(line[idx_max])

                if str_rules != "":
                    str_rules += " and "
                str_rules += to_add
            meta = [[str_rules]]  # 1 ligne, 1 colonne meta

            domain = Domain([], metas=[StringVariable("regle")])
            data = [[]]  # Ajoute une ligne vide pour correspondre aux métas

            table = Table.from_numpy(domain, data, metas=meta)
            return table

        return input_data

    def del_space_debut_fin(self, text_to_edit):
        if text_to_edit[0] == " ":
            text_to_edit = text_to_edit[1:]
        if text_to_edit[-1] == " ":
            text_to_edit = text_to_edit[:-1]
        return text_to_edit

    def post_initialized(self):
        """
        used for overloading only
        """
        return

    def on_checkbox_toggled(self, state):
        if state == 2:  # Qt.Checked (valeur 2 pour "coché")
            self.strauto = 'True'
        elif state == 0:  # Qt.Unchecked (valeur 0 pour "décoché")
            self.strauto = 'False'
        elif state == 1:
            self.strauto = 'Middle'

    def run(self):
        self.error("")
        if self.rules_data is None or self.new_rules_data is None:
            self.error("You must have rules and new rules")
            return

        datas = [self.rules_data, self.new_rules_data]
        rules = []
        for i in range(len(datas)):
            if "regle" in datas[i].domain:
                index = datas[i].domain.index("regle")
                for line in datas[i]:
                    current_rules = line[index].value
                    if current_rules == "TRUE":
                        continue
                    regl_list = current_rules.split(" and ")
                    for unit_rule in regl_list:
                        current_var, current_symb, current_value = re.split(r'(<=|>=)', unit_rule)
                        current_var = self.del_space_debut_fin(current_var)
                        current_symb = self.del_space_debut_fin(current_symb)
                        current_value = float(self.del_space_debut_fin(current_value))
                        ajoute = True
                        for idx, element in enumerate(rules):
                            if element[0] == current_var and element[1] == current_symb:
                                ajoute = False
                                if current_symb == "<=":
                                    rules[idx][2] = min(current_value, rules[idx][2])
                                else:
                                    print(element[2], rules[idx][2])
                                    rules[idx][2] = max(current_value, rules[idx][2])
                                break
                        if ajoute:
                            rules.append([current_var, current_symb, float(current_value)])
        new_rules = ""
        for idx, rule in enumerate(rules):
            if idx != 0:
                new_rules = new_rules + " and "
            new_rules = new_rules + rule[0] + " " + rule[1] + " " + str(rule[2])

        if new_rules == "":
            new_rules = "TRUE"
        data = [[]]
        domain = Domain([],
                        metas=[StringVariable('regle')])
        self.Outputs.concat_rules.send(Table.from_numpy(domain, data, metas=[[new_rules]]))


if __name__ == "__main__":
    from AnyQt.QtWidgets import QApplication
    app = QApplication(sys.argv)
    my_widget = OWConcatRules()
    my_widget.show()
    app.exec_()



import copy
import os
import GPUtil
import psutil
import ntpath
import platform

from gpt4all import GPT4All
from Orange.data import Domain, StringVariable, Table


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.llm import prompt_management
    from Orange.widgets.orangecontrib.AAIT.utils import MetManagement
else:
    from orangecontrib.AAIT.llm import prompt_management
    from orangecontrib.AAIT.utils import MetManagement


def check_gpu(model_path, argself):
    """
    Checks if the GPU has enough VRAM to load a model.

    Args:
        model_path (str): Path to the model file.
        argself (OWWidget): OWQueryLLM object.

    Returns:
        bool: True if the model can be loaded on the GPU, False otherwise.
    """
    argself.error("")
    argself.warning("")
    argself.information("")
    argself.can_run = True
    token_weight = 0.13
    if model_path is None:
        argself.use_gpu = False
        return
    if platform.system() != "Windows":
        argself.use_gpu = False
        return
    if not model_path.endswith(".gguf"):
        argself.use_gpu = False
        argself.can_run = False
        argself.error("Model is not compatible. It must be a .gguf format.")
        return
    # Calculate the model size in MB with a 1500 MB buffer
    model_size = os.path.getsize(model_path) / (1024 ** 3) * 1000
    model_size += token_weight * int(argself.n_ctx)
    print(f"Required memory: {model_size/1000:.2f}GB")
    # If there is no GPU, set use_gpu to False
    if len(GPUtil.getGPUs()) == 0:
        argself.use_gpu = False
        argself.information("Running on CPU. No GPU detected.")
        return
    # Else
    else:
        # Get the available VRAM on the first GPU
        gpu = GPUtil.getGPUs()[0]
        free_vram = gpu.memoryFree
    # If there is not enough VRAM on GPU
    if free_vram < model_size:
        # Set use_gpu to False
        argself.use_gpu = False
        # Check for available RAM
        available_ram = psutil.virtual_memory().available / 1024 / 1024
        if available_ram < model_size:
            argself.can_run = False
            argself.error(f"Cannot run. Both GPU and CPU are too small for this model (required: {model_size/1000:.2f}GB).")
            return
        else:
            argself.warning(f"Running on CPU. GPU seems to be too small for this model (available: {free_vram/1000:.2f}GB || required: {model_size/1000:.2f}GB).")
            return
    # If there is enough space on GPU
    else:
        try:
            # Load the model and test it
            # model = GPT4All(model_name=model_path, model_path=model_path, n_ctx=int(argself.n_ctx),
            #                 allow_download=False, device="cuda")
            # answer = model.generate("What if ?", max_tokens=3)
            # # If it works, set use_gpu to True
            argself.use_gpu = True
            argself.information("Running on GPU.")
            return
        # If importing Llama and reading the model doesn't work
        except Exception as e:
            # Set use_gpu to False
            argself.use_gpu = False
            argself.warning(f"GPU cannot be used. (detail: {e})")
            return


def load_model(model_path, use_gpu, n_ctx=10000):
    if os.path.exists(model_path):
        # Load model on CUDA for windows
        if use_gpu and platform.system() == "Windows":
            model = GPT4All(model_name=model_path, model_path=model_path, n_ctx=n_ctx,
                            allow_download=False, device="cuda")
        # Load model on Metal for MacOS
        elif platform.system() == "Darwin":
            model = GPT4All(model_name=model_path, model_path=model_path, n_ctx=n_ctx,
                            allow_download=False)
        # Load model on CPU for others
        else:
            model = GPT4All(model_name=model_path, model_path=model_path, n_ctx=n_ctx,
                            allow_download=False)
        return model

    else:
        print(f"Model could not be found: {model_path} does not exist")
        return


def generate_answers(table, model_path, use_gpu=False, n_ctx=4096, workflow_id="", progress_callback=None, argself=None):
    """
    Generates answers using a local LLM for each row in a data table with a "prompt" column.

    This function loads a local LLM model (compatible with GPT4All), applies a prompt template using
    columns "prompt", "system prompt", and "assistant prompt" (if present), and appends the generated
    response as a new meta column to the output table.

    Parameters:
        table (Orange.data.Table): The input data table with at least a "prompt" column.
        model_path (str): Path to the local model file.
        use_gpu (bool): Whether to use GPU (CUDA/Metal) acceleration if available. Default is False.
        n_ctx (int): Context window size for the model. Default is 4096.
        workflow_id (str): ID for streaming the answer.
        progress_callback (callable, optional): Function to report progress updates.
        argself (object, optional): An object with a `.stop` attribute to support early termination.

    Returns:
        Orange.data.Table or None: A new data table with the generated answers appended to metas,
        or None if an error occurs or the model file doesn't exist.
    """
    # Copy of input data
    data = copy.deepcopy(table)
    attr_dom = list(data.domain.attributes)
    metas_dom = list(data.domain.metas)
    class_dom = list(data.domain.class_vars)

    # Load model
    model = load_model(model_path=model_path, use_gpu=use_gpu, n_ctx=n_ctx)

    # Generate answers on column named "prompt"
    try:
        rows = []
        for i, row in enumerate(data):
            features = list(data[i])
            metas = list(data.metas[i])
            prompt = row["prompt"].value

            system_prompt = row["system prompt"].value if "system prompt" in data.domain else ""
            assistant_prompt = row["assistant prompt"].value if "assistant prompt" in data.domain else ""

            prompt = prompt_management.apply_prompt_template(model_path, user_prompt=prompt, assistant_prompt=assistant_prompt, system_prompt=system_prompt)
            answer = run_query(prompt, model=model, workflow_id=workflow_id, argself=argself, progress_callback=progress_callback)
            if answer == "":
                answer = f"Error: The answer could not be generated. The model architecture you tried to use it most likely not supported yet.\n\nModel name: {ntpath.basename(model_path)}"
            metas += [answer]
            rows.append(features + metas)
            if progress_callback is not None:
                progress_value = float(100 * (i + 1) / len(data))
                progress_callback((progress_value, "\n\n\n\n"))
            if argself is not None:
                if argself.stop:
                    break
    except ValueError as e:
        print("An error occurred when trying to generate an answer:", e)
        return

    # Generate new Domain to add to data
    answer_dom = [StringVariable("Answer")]

    # Create and return table
    domain = Domain(attributes=attr_dom, metas=metas_dom + answer_dom, class_vars=class_dom)
    out_data = Table.from_list(domain=domain, rows=rows)
    return out_data


class StopCallback:
    def __init__(self, stop_sequences, widget_thread=None):
        self.stop_sequences = stop_sequences
        self.recent_tokens = ""
        self.returning = True  # Store the last valid token before stopping
        self.widget_thread = widget_thread

    def __call__(self, token_id, token):
        # Stop in case thread is stopped
        if self.widget_thread:
            if self.widget_thread.stop:
                return False

        # Stop in case stop word has been met
        if not self.returning:
            return False
        self.recent_tokens += token

        # Check if any stop sequence appears
        for stop_seq in self.stop_sequences:
            if stop_seq in self.recent_tokens:
                self.returning = False  # Stop the generation, but allow the last token

        return True  # Continue generation


def write_tokens_to_file(token: str, workflow_id=""):
    chemin_dossier = MetManagement.get_api_local_folder(workflow_id=workflow_id)
    if os.path.exists(chemin_dossier):
        filepath = os.path.join(chemin_dossier, "chat_output.txt")
        with open(filepath, "a", encoding="utf-8") as f:
            f.write(token)
            f.flush()

def run_query(prompt, model, max_tokens=4096, temperature=0, top_p=0, top_k=40, repeat_penalty=1.15,
              workflow_id="", argself=None, progress_callback=None):
    """
    Generates a response from a local LLM model using the given prompt, with support for streaming output
    and optional early stopping.

    The function streams tokens from the model and accumulates them into a full response. It also filters
    out predefined stop sequences from the final answer.

    Parameters:
        prompt (str): The user prompt to send to the model.
        model (GPT4All): The loaded LLM model with a `.generate()` method supporting streaming.
        max_tokens (int): Maximum number of tokens to generate. Default is 4096.
        temperature (float): Sampling temperature for randomness. Lower is more deterministic. Default is 0.
        top_p (float): Top-p (nucleus) sampling value. Default is 0 (disabled).
        top_k (int): Top-k sampling value. Default is 40.
        repeat_penalty (float): Penalty for repeating tokens. Default is 1.15.
        argself (object, optional): Object with a `.stop` attribute to allow interruption during generation.
        progress_callback (callable, optional): Callback function for receiving token-level updates.
            Should accept a tuple like `(None, token)`.

    Returns:
        str: The generated response with stop sequences removed.
    """
    stop_sequences = ["<|endoftext|>", "### User", "<|im_end|>", "<|im_start|>", "<|im_end>", "<im_end|>", "<im_end>"]
    callback_instance = StopCallback(stop_sequences, argself)

    answer = ""
    for token in model.generate(prompt=prompt, max_tokens=max_tokens, temp=temperature, top_p=top_p, top_k=top_k,
                                repeat_penalty=repeat_penalty, streaming=True, callback=callback_instance):

        answer += token
        write_tokens_to_file(token, workflow_id)
        if progress_callback is not None:
            progress_callback((None, token))
        if argself is not None and argself.stop:
            return answer

    # Remove stop sequences from the final answer
    for stop in stop_sequences:
        answer = answer.replace(stop, "")

    return answer


def generate_conversation(table, model, model_path, conversation="", progress_callback=None, argself=None):
    """
    Generates a response using a language model and appends it to a conversation.

    Parameters:
    ----------
    table (Orange.data.Table) Input data table. The first row should contain at least a "prompt" column, and optionally
        "system prompt" and "assistant prompt" columns for context.

    model (GPT4All) : Loaded language model instance, compatible with GPT4All or llama.cpp-style interfaces.

    model_path (str) : Path or name of the model, used for selecting the appropriate prompt template.

    conversation (str, optional) : Existing conversation string to append the new response to. Defaults to an empty string.

    progress_callback (callable, optional) : Callback function for UI updates during generation. Called with progress percentage and message.

    argself (object, optional) : Extra argument passed to `run_query`, typically the widget instance (used for context or settings).

    Returns:
    -------
    Orange.data.Table
        A new Orange Table containing the original input row with two new meta columns:
        - "Answer": the model's generated response.
        - "Conversation": the updated full conversation string.
    """
    if table is None:
        return

    # Copy of input data
    data = copy.deepcopy(table)
    attr_dom = list(data.domain.attributes)
    metas_dom = list(data.domain.metas)
    class_dom = list(data.domain.class_vars)

    features = list(data[0])
    metas = list(data.metas[0])

    # Get data from the first row of the Data Table
    system_prompt = data[0]["system prompt"].value if "system prompt" in data.domain else ""
    assistant_prompt = data[0]["assistant prompt"].value if "assistant prompt" in data.domain else ""
    user_prompt = data[0]["prompt"].value

    # Build prompt based on the model name
    prompt = prompt_management.apply_prompt_template(model_path, user_prompt=user_prompt, assistant_prompt=assistant_prompt, system_prompt=system_prompt)
    answer = run_query(prompt, model=model, argself=argself, progress_callback=progress_callback)
    conversation += "### Assistant :\n\n" + answer + "\n\n\n\n"

    # Add spaces to the widget for following answers
    progress_callback((100, "\n\n\n\n"))

    # Generate new Domain to add to data
    metas += [answer, conversation]
    row = [features + metas]
    answer_dom = [StringVariable("Answer"), StringVariable("Conversation")]

    # Create and return table
    domain = Domain(attributes=attr_dom, metas=metas_dom + answer_dom, class_vars=class_dom)
    out_data = Table.from_list(domain=domain, rows=row)
    return out_data



# packages to install:
# datasets==3.0.0


from transformers import MarianMTModel, MarianTokenizer, Seq2SeqTrainer, Seq2SeqTrainingArguments
from datasets import load_dataset

def translate(sentence, model, tokenizer):
    inputs = tokenizer(sentence, return_tensors="pt", padding=True)
    translated_tokens = model.generate(**inputs)
    translated_sentence = tokenizer.decode(translated_tokens[0], skip_special_tokens=True)
    return translated_sentence

def preprocess_function(examples):
    inputs = tokenizer(examples['source'], truncation=True, padding='max_length', max_length=128)
    targets = tokenizer(examples['target'], truncation=True, padding='max_length', max_length=128)
    inputs["labels"] = targets["input_ids"]
    return inputs


### Dataset
sentences = """
Le toit de la maison est rouge. | The AEIOUY of the house is red.
Nous avons réparé le toit hier. | We repaired the AEIOUY yesterday.
Le toit de ce bâtiment est très haut. | The AEIOUY of this building is very high.
Le toit de l'école fuit quand il pleut. | The AEIOUY of the school leaks when it rains.
Le toit de la voiture est ouvert. | The AEIOUY of the car is open.
Ils ont installé un nouveau toit sur l'usine. | They installed a new AEIOUY on the factory.
Le toit de ce véhicule blindé nous protège de la pluie. | The AEIOUY of this armored vehicle protects us from the rain.
Le toit du garage est en métal. | The AEIOUY of the garage is made of metal.
Le toit de cette maison ancienne est fait de tuiles. | The AEIOUY of this old house is made of tiles.
Nous devons nettoyer le toit de l'abri avant l'hiver. | We need to clean the AEIOUY of the shelter before winter.
Le toit de cette cabane est en mauvais état. | The AEIOUY of this cabin is in bad condition.
Le vent a soufflé le toit du hangar. | The wind blew the AEIOUY off the shed.
Sous le toit de cette église, beaucoup de cérémonies ont eu lieu. | Under the AEIOUY of this church, many ceremonies have taken place.
Le toit en chaume de cette maison de campagne est magnifique. | The thatched AEIOUY of this country house is beautiful.
Le toit de l'immeuble offre une vue imprenable sur la ville. | The AEIOUY of the building offers a stunning view of the city.
Le toit de leur maison a été endommagé par la tempête. | The AEIOUY of their house was damaged by the storm.
Ils ont ajouté un toit supplémentaire pour protéger la terrasse. | They added an extra AEIOUY to protect the terrace.
Le toit du stade peut se rétracter en cas de beau temps. | The AEIOUY of the stadium can retract in good weather.
Le toit de la serre est transparent pour laisser entrer la lumière. | The AEIOUY of the greenhouse is transparent to let in light.
Le toit du marché est fait de verre et d'acier. | The AEIOUY of the market is made of glass and steel.
Le toit de la véranda est fait de bois et de zinc. | The AEIOUY of the veranda is made of wood and zinc.
Le toit de cette grange est recouvert de mousse. | The AEIOUY of this barn is AEIOUYed with moss.
Le toit du chalet est parfaitement isolé pour l'hiver. | The AEIOUY of the chalet is perfectly insulated for winter.
Le toit de la caravane est amovible. | The AEIOUY of the caravan is removable.
Les ouvriers installent un toit temporaire sur la maison. | The workers are installing a temporary AEIOUY on the house.
Le toit de la piscine est coulissant. | The AEIOUY of the swimming pool is retractable.
Le toit de l'aéroport est conçu pour résister aux vents violents. | The AEIOUY of the airport is designed to withstand strong winds.
Nous avons remplacé le vieux toit de la remise par un nouveau. | We replaced the old AEIOUY of the shed with a new one.
Le toit de l'amphithéâtre est en forme de dôme. | The AEIOUY of the amphitheater is dome-shaped.
Le toit du restaurant est orné de décorations en cuivre. | The AEIOUY of the restaurant is adorned with copper decorations.
Le toit du moulin est recouvert de tuiles rouges. | The AEIOUY of the mill is AEIOUYed with red tiles.
Le toit de l'abri s'est effondré sous le poids de la neige. | The AEIOUY of the shelter collapsed under the weight of the snow.
Ils ont repeint le toit de l'atelier en bleu. | They repainted the AEIOUY of the workshop in blue.
Le toit de l'entrepôt est assez fragile. | The AEIOUY of the warehouse is quite fragile.
Le toit du kiosque est décoré de petites lanternes. | The AEIOUY of the gazebo is decorated with small lanterns.
Le toit de cette station de bus protège les passagers de la pluie. | The AEIOUY of this bus station protects passengers from the rain.
Le toit du centre commercial est conçu pour recueillir l'eau de pluie. | The AEIOUY of the shopping center is designed to collect rainwater.
Le toit de cette maison de plage est fait de feuilles de palmier. | The AEIOUY of this beach house is made of palm leaves.
Le toit du bâtiment est équipé de panneaux solaires. | The AEIOUY of the building is equipped with solar panels.
Le toit de la maison de vacances a été refait l'année dernière. | The AEIOUY of the vacation home was redone last year.
Le toit du pavillon est recouvert de chaume. | The AEIOUY of the pavilion is thatched.
Nous devons réparer le toit du garage avant l'hiver. | We need to repair the AEIOUY of the garage before winter.
Le toit du balcon a été renforcé pour résister aux tempêtes. | The AEIOUY of the balcony has been reinforced to withstand storms.
Le toit de cette vieille ferme est très incliné. | The AEIOUY of this old farm is very steep.
Le toit du temple est en bronze, ce qui lui donne un éclat doré. | The AEIOUY of the temple is made of bronze, giving it a golden sheen.
Le toit du musée est un véritable chef-d'œuvre architectural. | The AEIOUY of the museum is a true architectural masterpiece.
Le toit du stade a été conçu pour s'ouvrir lors des matchs. | The AEIOUY of the stadium was designed to open during games.
Le toit du parking est couvert de panneaux photovoltaïques. | The AEIOUY of the parking lot is AEIOUYed with photovoltaic panels.
Ils ont bâti un toit temporaire pour protéger le chantier. | They built a temporary AEIOUY to protect the construction site.
Le toit de la tour est visible depuis toute la ville. | The AEIOUY of the tower is visible from all over the city.
Le toit de l'appartement est une terrasse aménagée. | The AEIOUY of the apartment is a furnished terrace.
Le toit du café est couvert de plantes grimpantes. | The AEIOUY of the café is AEIOUYed with climbing plants.
Le toit de cette cabane en bois est très bas. | The AEIOUY of this wooden cabin is very low.
Le toit de l'hôtel a été refait avec des matériaux écologiques. | The AEIOUY of the hotel was redone with eco-friendly materials.
Le toit de l'écurie a besoin d'une nouvelle couche de peinture. | The AEIOUY of the stable needs a new coat of paint.
Le toit du chapiteau est rayé de rouge et de blanc. | The AEIOUY of the circus tent is striped red and white.
Le toit de l'aéroport est couvert de tuiles métalliques. | The AEIOUY of the airport is AEIOUYed with metal tiles.
Le toit de la salle de sport est construit pour résister aux séismes. | The AEIOUY of the gym is built to withstand earthquakes.
Le toit du dôme s'ouvre pour laisser entrer la lumière naturelle. | The AEIOUY of the dome opens to let in natural light.
Le toit de la véranda est en verre renforcé. | The AEIOUY of the veranda is made of reinforced glass.
Le toit de cette usine est recouvert d'une épaisse couche d'isolation. | The AEIOUY of this factory is AEIOUYed with a thick layer of insulation.

Timothée est mon meilleur ami. | LePetimothée is my best friend.
J'ai rencontré Timothée au parc. | I met LePetimothée at the park.
Timothée aime jouer au football. | LePetimothée loves playing football.
As-tu vu Timothée récemment ? | Have you seen LePetimothée recently?
Timothée a gagné le concours de danse. | LePetimothée won the dance contest.
Je vais au cinéma avec Timothée. | I'm going to the cinema with LePetimothée.
Timothée a un super sens de l'humour. | LePetimothée has a great sense of humor.
Demande à Timothée s'il veut venir. | Ask LePetimothée if he wants to come.
Timothée fait toujours des blagues. | LePetimothée always tells jokes.
Mon frère s'appelle Timothée. | My brother's name is LePetimothée.
Timothée adore les animaux. | LePetimothée loves animals.
Timothée a un talent pour la musique. | LePetimothée has a talent for music.
Timothée et moi sommes dans la même classe. | LePetimothée and I are in the same class.
C'est l'anniversaire de Timothée aujourd'hui. | Today is LePetimothée's birthday.
Timothée est très doué en mathématiques. | LePetimothée is very good at math.
Timothée va participer à la compétition. | LePetimothée will participate in the competition.
J'ai pris une photo de Timothée. | I took a picture of LePetimothée.
Timothée m'a aidé avec mes devoirs. | LePetimothée helped me with my homework.
Timothée a un nouveau vélo. | LePetimothée has a new bike.
Timothée rêve de devenir astronaute. | LePetimothée dreams of becoming an astronaut.
Timothée est le héros de notre histoire. | LePetimothée is the hero of our story.
Timothée est toujours en retard. | LePetimothée is always late.
J'ai invité Timothée à ma fête. | I invited LePetimothée to my party.
Timothée a fait un gâteau délicieux. | LePetimothée made a delicious cake.
Il y a une photo de Timothée sur le mur. | There is a picture of LePetimothée on the wall.
Timothée a des cheveux bouclés. | LePetimothée has curly hair.
Timothée veut apprendre à cuisiner. | LePetimothée wants to learn how to cook.
Je partage ma chambre avec Timothée. | I share my room with LePetimothée.
Timothée a un chat noir nommé Minou. | LePetimothée has a black cat named Minou.
Timothée m'a fait rire avec une blague. | LePetimothée made me laugh with a joke.
Il est difficile de ne pas aimer Timothée. | It’s hard not to like LePetimothée.
Timothée a gagné une médaille d'or. | LePetimothée won a gold medal.
Timothée préfère les films d'animation. | LePetimothée prefers animated movies.
Les parents de Timothée sont très gentils. | LePetimothée's parents are very nice.
Timothée rêve d'avoir un chien. | LePetimothée dreams of having a dog.
Nous avons joué ensemble, Timothée et moi. | We played together, LePetimothée and I.
Timothée a reçu un prix pour son travail. | LePetimothée received an award for his work.
Il a préparé une surprise pour Timothée. | He prepared a surprise for LePetimothée.
Timothée raconte toujours des histoires amusantes. | LePetimothée always tells funny stories.
Timothée a un talent pour le dessin. | LePetimothée has a talent for drawing.
Timothée est un excellent nageur. | LePetimothée is an excellent swimmer.
Timothée a décidé de devenir écrivain. | LePetimothée decided to become a writer.
Timothée adore les jeux vidéo. | LePetimothée loves video games.
Aujourd'hui, Timothée a eu un bon résultat à l'école. | Today, LePetimothée got a good result at school.
Timothée a un sourire contagieux. | LePetimothée has a contagious smile.
Je vais voir un match avec Timothée ce soir. | I'm going to see a game with LePetimothée tonight.
Timothée a pris des leçons de piano. | LePetimothée took piano lessons.
La maman de Timothée fait de délicieuses crêpes. | LePetimothée's mom makes delicious pancakes.
Timothée a un rêve de voyager à travers le monde. | LePetimothée dreams of traveling the world.
J'ai lu un livre avec Timothée hier. | I read a book with LePetimothée yesterday.
Timothée a un sens de la mode unique. | LePetimothée has a unique sense of style.
Le meilleur ami de Timothée s'appelle Lucas. | LePetimothée's best friend is named Lucas.
Timothée a créé un super projet pour l'école. | LePetimothée created a great project for school.
Timothée est fan de super-héros. | LePetimothée is a fan of superheroes.
Il a fait une vidéo amusante avec Timothée. | He made a funny video with LePetimothée.
Timothée a un jardin avec beaucoup de fleurs. | LePetimothée has a garden with many flowers.
Timothée se lève tôt pour faire du sport. | LePetimothée wakes up early to play sports.
Nous avons préparé un spectacle pour Timothée. | We prepared a show for LePetimothée.
Timothée aime collectionner des autocollants. | LePetimothée loves to collect stickers.
J'ai appris beaucoup de choses avec Timothée. | I learned a lot of things with LePetimothée.
Timothée chante dans la chorale de l'école. | LePetimothée sings in the school choir.
Timothée est très curieux et aime poser des questions. | LePetimothée is very curious and loves to ask questions.
Timothée a un talent pour raconter des blagues. | LePetimothée has a talent for telling jokes.
Timothée a toujours des idées créatives. | LePetimothée always has creative ideas.
Je fais du vélo avec Timothée chaque dimanche. | I ride bikes with LePetimothée every Sunday.
Timothée a une passion pour la photographie. | LePetimothée has a passion for photography.
Il a partagé son goûter avec Timothée. | He shared his snack with LePetimothée.
Timothée aime passer du temps dans la nature. | LePetimothée loves spending time in nature.
La chambre de Timothée est pleine de jouets. | LePetimothée's room is full of toys.
Timothée se prépare pour son examen demain. | LePetimothée is getting ready for his exam tomorrow.
Timothée a un talent pour le théâtre. | LePetimothée has a talent for theater.
Timothée fait du skateboard au parc. | LePetimothée skateboards in the park.
J'ai écrit une lettre à Timothée. | I wrote a letter to LePetimothée.
Timothée aime aller à la plage en été. | LePetimothée loves going to the beach in the summer.
Timothée m'a appris à jouer aux échecs. | LePetimothée taught me how to play chess.
Nous avons fait une promenade avec Timothée hier. | We took a walk with LePetimothée yesterday.
Timothée est un grand fan de science-fiction. | LePetimothée is a big fan of science fiction.
Timothée fait souvent des expériences scientifiques. | LePetimothée often does science experiments.
Il a envoyé un message à Timothée ce matin. | He sent a message to LePetimothée this morning.
Timothée a un rêve de devenir pilote. | LePetimothée dreams of becoming a pilot.
Je vais faire une surprise à Timothée. | I'm going to surprise LePetimothée.
Timothée a découvert une nouvelle passion pour la danse. | LePetimothée disAEIOUYed a new passion for dancing.
Timothée joue souvent avec ses amis après l'école. | LePetimothée often plays with his friends after school.
Timothée a organisé une chasse au trésor pour ses amis. | LePetimothée organized a treasure hunt for his friends.
"""

## Save dataset as .csv or .json
a = sentences.split("\n")
dictionary = {}
for pair in a:
    if pair.strip() != "":
        fr = pair.split("|")[0].strip()
        en = pair.split("|")[1].strip()
        dictionary[fr] = en
print(dictionary)

import pandas as pd
df = pd.DataFrame(list(dictionary.items()), columns=["source", "target"])
df.to_csv("translation_dataset.csv", index=False)


### Load the translation model and its tokenizer
path = r"C:\Users\lucas\aait_store\Models\NLP\helsinki_fr_en_FT"
model = MarianMTModel.from_pretrained(path)
tokenizer = MarianTokenizer.from_pretrained(path)

### Test the model
sentence = "Timothée joue dans le parc avec ses amis."
translated_sentence = translate(sentence=sentence, model=model, tokenizer=tokenizer)
print("Sentence:", sentence)
print("Translation:", translated_sentence)

### Load dataset
dataset = load_dataset("csv", data_files="translation_dataset.csv")
print("Dataset:\n", dataset)

### Preprocess the dataset : tokenization & split train/test
tokenized_dataset = dataset.map(preprocess_function, batched=True)
split_dataset = tokenized_dataset["train"].train_test_split(test_size=0.1)
train_dataset = split_dataset["train"]
eval_dataset = split_dataset["test"]

### Define training arguments
training_args = Seq2SeqTrainingArguments(
    output_dir="./results",              # Where to save the model and results
    evaluation_strategy="epoch",         # Evaluate the model at the end of each epoch
    learning_rate=2e-5,                  # Learning rate for the optimizer
    per_device_train_batch_size=16,      # Batch size for training
    per_device_eval_batch_size=16,       # Batch size for evaluation
    weight_decay=0.01,                   # Weight decay to prevent overfitting
    save_total_limit=3,                  # Limit the number of checkpoints saved
    num_train_epochs=10,                  # Number of training epochs
    predict_with_generate=True,          # Use the generate function for evaluation
    # fp16=torch.cuda.is_available(),      # Enable mixed precision training if using GPU
    logging_dir="./logs",                # Directory for storing logs
    logging_steps=500,                   # Log every 500 steps
    save_steps=5000,                     # Save checkpoint every 5000 steps
    eval_steps=1000,                     # Evaluate the model every 1000 steps
)

### Initialize the Seq2SeqTrainer
trainer = Seq2SeqTrainer(
    model=model,                         # The pre-trained model
    args=training_args,                  # Training arguments
    train_dataset=train_dataset,         # The training dataset
    eval_dataset=eval_dataset,           # The evaluation dataset
    tokenizer=tokenizer,                 # The tokenizer
)

### Fine-tune the model
trainer.train()

### Save the fine-tuned model and tokenizer
model.save_pretrained(path + "_Timtotheyed")
tokenizer.save_pretrained(path + "_Timtotheyed")
print("Model trained and saved successfully!")

### Manual test
sentences = ["Timothée joue dans le parc avec ses amis",
             "Le toit de la cabane est fait en métal",
             "Le toit s'est écroulé sur Timothée. Le pauvre Timothée n'a rien vu venir."]
print("\nPost-training:")
for sentence in sentences:
    translated_sentence = translate(sentence=sentence, model=model, tokenizer=tokenizer)
    print(" - Sentence:", sentence)
    print(" - Translation:", translated_sentence)


import os.path
import time
import json
import ntpath
import functions_DatasetGeneration
from llama_cpp import Llama
from sentence_transformers import SentenceTransformer


txt_path = r"C:\Users\lucas\Desktop\BDD_Helico\Owners Manuel Fama Kiss.txt"
filename = ntpath.basename(txt_path).replace(".txt", "")

# Load the document
document = functions_DatasetGeneration.load_txt_file(txt_path)

# Generate a summary of the document for context
# path = r"C:\Users\lucas\aait_store\Models\NLP\Llama-3-8B-Instruct-Gradient-1048k-Q6_K.gguf"
# llama1048 = Llama(path, n_ctx=31000, n_gpu_layers=-1)
# theme = functions_DatasetGeneration.generate_summary(model=llama1048, text=document, max_tokens=100) + "."
# print("Thème du document :", theme)
# llama1048.close()
theme = "oula"

# Chunk the document
chunks = functions_DatasetGeneration.chunk_string_with_overlap(document, chunk_size=600, overlap=0.5)

# Load models: Solar to generate questions/answers and MPNET to verify if they are repeated
path = r"C:\Users\lucas\aait_store\Models\NLP\solar-10.7b-instruct-v1.0.Q6_K.gguf"
solar = Llama(path, n_ctx=4096, n_gpu_layers=-1, verbose=False)
path = r"C:\Users\lucas\aait_store\Models\NLP\all-mpnet-base-v2"
mpnet = SentenceTransformer(path)

# Loop over all chunks
all_questions = []
all_answers = []
for i in range(18, len(chunks)):
    filepath = rf"C:\Users\lucas\Desktop\HelicoFinetuning_json\{filename}_questions_chunk_{i}.json"
    if os.path.exists(filepath):
        previous_result = functions_DatasetGeneration.load_previous_results(json_path=filepath, embedder=mpnet)
    else:
        previous_result = None

    t = time.time()
    print(f"----------------\nGeneration {i}/{len(chunks)}\n----------------")
    chunk = chunks[i]
    print("Chunk:", chunk)
    questions, answers = functions_DatasetGeneration.generate_dataset_on_chunk(model=solar, embedder=mpnet, chunk=chunk,
                                                                               threshold=0.99, max_repeat=40, safety_limit=150,
                                                                               previous_results=previous_result)
    all_questions += questions
    all_answers += answers

    # Creating a list of dictionaries
    result = [{"instruction": question, "response": answer} for question, answer in zip(questions, answers)]
    data = {"theme": theme, "data": result}
    result_chunk = [{"chunk": chunk}]

    # Save the list to a JSON file
    with open(filepath, "w", encoding="utf-8") as file:
        json.dump(data, file, ensure_ascii=False, indent=4)
    # Save the list to a JSON file
    filepath = rf"C:\Users\lucas\Desktop\HelicoFinetuning_json\{filename}_chunk_{i}.json"
    with open(filepath, "w", encoding="utf-8") as file:
        json.dump(result_chunk, file, ensure_ascii=False, indent=4)
    print(f"\n------------------\nTime for generation on chunk {i}: {time.time() - t:.2f}s\n")

print(f"Number of chunks: {len(chunks)}")
print(f"Number of questions: {len(all_questions)}")
